import collections.abc
import pathlib
import typing
import sys

import numpy
import numpy.typing
from scipy import integrate
from scipy.interpolate import interp1d

from .. import aliased
from .. import container
from .. import datafile
from .. import etc
from .. import measured
from .. import metric
from .. import numeric
from .. import parameter
from .. import physical
from .. import quantity
from .. import reference
from .. import symbolic
from .. import universal
from ..reference import ARRAYS
from ._objects import (
    Arguments,
    Context,
    Quantity,
)


class Implementations:
    """Functional forms of all observable EPREM quantities."""

    def __init__(
        self,
        dataset: datafile.Arrays,
        axes: datafile.Axes,
        grid: aliased.Mapping[str, datafile.Array],
        constants: universal.Constants,
        parameters: parameter.Interface,
        system: metric.System,
    ) -> None:
        self._dataset = dataset
        self._axes = axes
        self._grid = grid
        self._constants = constants
        self._parameters = parameters
        self._system = system
        self._definitions = None

    @property
    def definitions(self):
        """A mapping from quantity name to canonical method."""
        if self._definitions is None:
            self._definitions = {
                'time': self.time,
                'energy': self.energy,
                'v': self.v,
                'mu': self.mu,
                'mass': self.mass,
                'charge': self.charge,
                'r': self.r,
                'theta': self.theta,
                'phi': self.phi,
                'br': self.br,
                'btheta': self.btheta,
                'bphi': self.bphi,
                'ur': self.ur,
                'utheta': self.utheta,
                'uphi': self.uphi,
                'rho': self.rho,
                'f': self.f,
                'x': self.x,
                'y': self.y,
                'z': self.z,
                'bmag': self.bmag,
                'umag': self.umag,
                'ubmag': self.ubmag,
                'angle': self.angle,
                'upara': self.upara,
                'udotb': self.udotb,
                'uperp': self.uperp,
                'divu': self.divu,
                'rigidity': self.rigidity,
                'mfp': self.mfp,
                'ar': self.ar,
                'average energy': self.average_energy,
                'energy density': self.energy_density,
                'flux': self.flux,
                'fluence': self.fluence,
                'intflux': self.intflux,
            }
        return self._definitions

    ## -- auxiliary/axis-like quantities --
    # Each method directly subscripts the dataset array because interpolation
    # over any of these quantities is either meaningless or unnecessary.

    def time(self, time=None) -> physical.Array:
        """Compute the output times."""
        return self.dataset.time[_normalize_indices(time or slice(None))]

    def energy(self, energy=None) -> physical.Array:
        """Compute the particle energies."""
        return self.dataset.energy[_normalize_indices(energy or slice(None))]

    def v(self, energy=None) -> physical.Array:
        """Compute the particle speeds."""
        return self.dataset.v[_normalize_indices(energy or slice(None))]

    def mu(self, mu=None) -> physical.Array:
        """Compute the particle pitch-angle cosines."""
        return self.dataset.mu[_normalize_indices(mu or slice(None))]

    def mass(self, species=None) -> physical.Array:
        """Compute the mass of each species."""
        return self.dataset.mass[_normalize_indices(species or slice(None))]

    def charge(self, species=None) -> physical.Array:
        """Compute the charge of each species."""
        return self.dataset.charge[_normalize_indices(species or slice(None))]

    ## -- primary observable quantities --
    # Each method accepts only the indices corresponding to the dataset
    # dimensions because it does not need to call other methods. It also accepts
    # arbitrary assumptions in case the user wants to interpolate over any of
    # the dataset dimensions.

    def r(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the radial observer coordinates."""
        return self._observe(
            self.dataset.r,
            assumptions,
            time=time,
            shell=shell,
        )

    def theta(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the polar observer coordinates."""
        return self._observe(
            self.dataset.theta,
            assumptions,
            time=time,
            shell=shell,
        )

    def phi(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the azimuthal observer coordinates."""
        return self._observe(
            self.dataset.phi,
            assumptions,
            time=time,
            shell=shell,
        )

    def br(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the magnetic-field radial component."""
        return self._observe(
            self.dataset.br,
            assumptions,
            time=time,
            shell=shell,
        )

    def btheta(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the magnetic-field polar component."""
        return self._observe(
            self.dataset.btheta,
            assumptions,
            time=time,
            shell=shell,
        )

    def bphi(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the magnetic-field azimuthal component."""
        return self._observe(
            self.dataset.bphi,
            assumptions,
            time=time,
            shell=shell,
        )

    def ur(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the velocity-field radial component."""
        return self._observe(
            self.dataset.ur,
            assumptions,
            time=time,
            shell=shell,
        )

    def utheta(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the velocity-field polar component."""
        return self._observe(
            self.dataset.utheta,
            assumptions,
            time=time,
            shell=shell,
        )

    def uphi(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the velocity-field azimuthal component."""
        return self._observe(
            self.dataset.uphi,
            assumptions,
            time=time,
            shell=shell,
        )

    def rho(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the plasma density."""
        return self._observe(
            self.dataset.rho,
            assumptions,
            time=time,
            shell=shell,
        )

    def mfp(
        self,
        time=None,
        shell=None,
        species=None,
        energy=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the particle mean free path."""
        return self._observe(
            self.dataset.mfp,
            assumptions,
            time=time,
            shell=shell,
            species=species,
            energy=energy,
        )

    def f(
        self,
        time=None,
        shell=None,
        species=None,
        energy=None,
        mu=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the full particle distribution."""
        return self._observe(
            self.dataset.f,
            assumptions,
            time=time,
            shell=shell,
            species=species,
            energy=energy,
            mu=mu,
        )

    ## -- derived observable quantities --
    # Each method accepts all the indices required by the intermediate
    # quantities on which it depends. It also accepts arbitrary assumptions,
    # which it passes along to intermediate quantities.

    def x(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the x coordinate."""
        r = self.r(time, shell, assumptions)
        theta = self.theta(time, shell, assumptions)
        phi = self.phi(time, shell, assumptions)
        # NOTE: Intermediate methods have already restricted the result via
        # subscription or interpolation.
        return r * numpy.sin(theta) * numpy.cos(phi)

    def y(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the y-axis coordinate."""
        r = self.r(time, shell, assumptions)
        theta = self.theta(time, shell, assumptions)
        phi = self.phi(time, shell, assumptions)
        # NOTE: Intermediate methods have already restricted the result via
        # subscription or interpolation.
        return r * numpy.sin(theta) * numpy.sin(phi)

    def z(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the z-axis coordinate."""
        r = self.r(time, shell, assumptions)
        theta = self.theta(time, shell, assumptions)
        # NOTE: Intermediate methods have already restricted the result via
        # subscription or interpolation.
        return r * numpy.cos(theta)

    def bmag(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the magnetic-field magnitude."""
        br = self.br(time, shell, assumptions)
        btheta = self.btheta(time, shell, assumptions)
        bphi = self.bphi(time, shell, assumptions)
        # NOTE: Intermediate methods have already restricted the result via
        # subscription or interpolation.
        return numpy.sqrt(br**2 + btheta**2 + bphi**2)

    def umag(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the velocity-field magnitude."""
        ur = self.ur(time, shell, assumptions)
        utheta = self.utheta(time, shell, assumptions)
        uphi = self.uphi(time, shell, assumptions)
        # NOTE: Intermediate methods have already restricted the result via
        # subscription or interpolation.
        return numpy.sqrt(ur**2 + utheta**2 + uphi**2)

    def ubmag(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute `umag * bmag`."""
        bmag = self.bmag(time, shell, assumptions)
        umag = self.umag(time, shell, assumptions)
        # NOTE: Intermediate methods have already restricted the result via
        # subscription or interpolation.
        return bmag * umag

    def angle(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the angle between U and B."""
        udotb = self.udotb(time, shell, assumptions)
        ubmag = self.ubmag(time, shell, assumptions)
        ratio = udotb / ubmag
        arg = numpy.array(ratio)
        arg[arg < -1.0] = -1.0
        arg[arg > +1.0] = +1.0
        # NOTE: Intermediate methods in `_compute_angle` have already restricted
        # the result via subscription or interpolation.
        return physical.array(
            numpy.arccos(arg),
            unit='rad',
            axes=self.dataset.ur.axes,
        )

    def uperp(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the perpendicular velocity component."""
        umag = self.umag(time, shell, assumptions)
        upara = self.upara(time, shell, assumptions)
        uperpsqr = umag**2 - upara**2
        # NOTE: Intermediate methods have already restricted the result via
        # subscription or interpolation.
        return numpy.sqrt(uperpsqr)

    def upara(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the parallel velocity component."""
        udotb = self.udotb(time, shell, assumptions)
        bmag = self.bmag(time, shell, assumptions)
        # NOTE: Intermediate methods have already restricted the result via
        # subscription or interpolation.
        return udotb / bmag

    def udotb(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute U·B."""
        br = self.br(time, shell, assumptions)
        ur = self.ur(time, shell, assumptions)
        btheta = self.btheta(time, shell, assumptions)
        utheta = self.utheta(time, shell, assumptions)
        bphi = self.bphi(time, shell, assumptions)
        uphi = self.uphi(time, shell, assumptions)
        # NOTE: Intermediate methods have already restricted the result via
        # subscription or interpolation.
        return br*ur + btheta*utheta + bphi*uphi

    def divu(
        self,
        time=None,
        shell=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the velocity divergence.

        The method computes ∇·U from the continuity equation, which can be
        written in the co-moving frame as

        ```
             1  dn
            --- -- = - div(U)
             n  dt
        ```

        where n represents density and d/dt is the convective derivative.
        """
        # NOTE: We first compute the entire array, then extract the subarray.
        rho = self.dataset.rho
        divu = -1.0 * numpy.gradient(rho, self.dataset.time, axis=0) / rho
        return self._observe(
            divu,
            assumptions,
            time=time,
            shell=shell,
        )

    def rigidity(
        self,
        species=None,
        energy=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the magnetic rigidity.

        This method computes the particle momentum based on the relativistic
        relation E² = (K + mc²)² = p²c² + (mc²)², where K is the particles'
        kinetic energy. Finally, it divides the momentum by the particle charge
        to get rigidity.
        """
        # NOTE: We "observe" energy instead of calling the corresponding method
        # in case we need to interpolate.
        return self._compute_rigidity(
            self.mass(species=species),
            self.charge(species=species),
            self._observe(self.dataset.energy, assumptions, energy=energy),
        )

    def _compute_rigidity(
        self,
        mass: physical.Array,
        charge: physical.Array,
        energy: physical.Array,
    ) -> physical.Array:
        """Internal computational logic for `_rigidity`.
        
        Notes
        -----
        - This method exists to support both `_rigidity`, which must have a
          function signature that conforms to what
          `observable.Context.implement` expects, and `_mfp`, which needs to be
          able to specify the energy argument.
        - This method requires that both `mass` and `scalar` be instances of
          `physical.Array`, which should always be true when they come from
          `self.mass` and `self.charge`.
        """
        if not isinstance(mass, physical.Array):
            raise TypeError(f"Wrong type for mass: {type(mass)}")
        if not isinstance(charge, physical.Array):
            raise TypeError(f"Wrong type for charge: {type(charge)}")
        c = self.constants['c'].asscalar
        j = energy / mass
        t = j.transpose('species', 'energy') + c**2
        p = numpy.sqrt(t**2 - c**4) * (mass / c)
        return p / charge

    def mfp_r(
        self,
        time=None,
        shell=None,
        species=None,
        energy=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the mean free path proportional to r."""
        # NOTE: We "observe" energy instead of calling the corresponding method
        # in case we need to interpolate.
        e = self._observe(self.dataset.energy, assumptions, energy=energy)
        m = self.mass(species=species)
        q = self.charge(species=species)
        rg = self._compute_rigidity(m, q, e)
        if not isinstance(e, physical.Array):
            raise TypeError(f"Wrong type for energy: {type(e)}")
        e0 = physical.array(
            [float(self.constants['MeV'])],
            unit=e.unit,
            axes=e.axes
        )
        rg0 = self._compute_rigidity(m, q, e0)
        r0 = self.constants['au'].asscalar
        r = self.r(time, shell, assumptions)
        rterm = (r / r0) ** self.parameters['mfpPower']
        rgterm = (rg / rg0) ** self.parameters['rigidityPower']
        lambda0 = self.parameters['lambda0'].withunit(str(self.system))
        # NOTE: Intermediate methods have already restricted the result via
        # subscription or interpolation.
        return lambda0 * rterm * rgterm

    def mfp_B(
        self,
        time=None,
        shell=None,
        species=None,
        energy=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the mean free path proportional to B."""
        # NOTE: We "observe" energy instead of calling the corresponding method
        # in case we need to interpolate.
        e = self._observe(self.dataset.energy, assumptions, energy=energy)
        m = self.mass(species=species)
        q = self.charge(species=species)
        rg = self._compute_rigidity(m, q, e)
        if not isinstance(e, physical.Array):
            raise TypeError(f"Wrong type for energy: {type(e)}")
        e0 = physical.array(
            [float(self.constants['MeV'])],
            unit=e.unit,
            axes=e.axes
        )
        rg0 = self._compute_rigidity(m, q, e0)
        b = self.bmag(time, shell, assumptions)
        b0 = self.parameters['mhdBAu'].withunit(b.unit)
        bterm = (b0 / b) ** (0.5*self.parameters['mfpPower'])
        rgterm = (rg / rg0) ** self.parameters['rigidityPower']
        lambda0 = self.parameters['lambda0'].withunit(str(self.system))
        # NOTE: Intermediate methods have already restricted the result via
        # subscription or interpolation.
        return lambda0 * bterm * rgterm

    def ar(
        self,
        time=None,
        shell=None,
        species=None,
        energy=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the acceleration rate."""
        # NOTE: This method does not attempt to subscript intermediate
        # quantities by time or energy in case downstream code needs to
        # interpolate over them. It subscripts by shell because it assumes that
        # upstream code has reset `shell` as necessary for radial interpolation,
        # and it subscripts by species because interpolation over species is
        # meaningless.
        ar = self._compute_ar(shell=shell, species=species)
        return self._observe(ar, assumptions, time=time, energy=energy)

    def _compute_ar(self, shell=None, species=None) -> physical.Array:
        """Helper for `_ar`."""
        s = (
            shell if isinstance(shell, numeric.index.Sequence)
            else self.axes['shell'].index()
        )
        s1 = s.shift(1, ceil=len(self.axes['shell'])-1)
        s2 = s.shift(-1, floor=0)
        ur1 = self.ur(shell=s1)
        utheta1 = self.utheta(shell=s1)
        uphi1 = self.uphi(shell=s1)
        ur2 = self.ur(shell=s2)
        utheta2 = self.utheta(shell=s2)
        uphi2 = self.uphi(shell=s2)
        ux1 = self.umag(shell=s1)
        u1dotu2 = ur1*ur2 + utheta1*utheta2 + uphi1*uphi2
        one = physical.scalar(1)
        delta_u = ux1 * abs(one - u1dotu2 / (ux1*ux1))
        kxx1_ux1 = self._kxx_ux(shell=s1, species=species)
        kxx2_ux2 = self._kxx_ux(shell=s2, species=species)
        # NOTE: We can't directly compute delta_x = kxx1_ux1 + kxx2_ux2 because
        # the kxx_ux arrays have different shell axes.
        data = numpy.array(kxx1_ux1) + numpy.array(kxx2_ux2)
        if kxx1_ux1.unit != kxx2_ux2.unit:
            raise ValueError(
                "Units mismatch while computing acceleration rate"
            ) from None
        unit = kxx1_ux1.unit
        if kxx1_ux1.dimensions != kxx2_ux2.dimensions:
            raise ValueError(
                "Dimensions mismatch while computing acceleration rate"
            ) from None
        keys = {d for d in kxx1_ux1.dimensions if d != 'shell'}
        if any(kxx1_ux1.axes[k] != kxx2_ux2.axes[k] for k in keys):
            raise ValueError(
                "Axes mismatch while computing acceleration rate"
            ) from None
        axes = {
            k: physical.points(s) if k == 'shell' else v
            for k, v in kxx1_ux1.axes.items()
        }
        delta_x = physical.array(data, unit=unit, axes=axes)
        return delta_u / delta_x

    def _kxx_ux(self, **axes):
        """Helper for acceleration rate."""
        t = axes.get('time')
        s = axes.get('shell')
        p = axes.get('species')
        e = axes.get('energy')
        v = self.v(energy=e)
        mfp = self.mfp(time=t, shell=s, species=p, energy=e)
        theta = self.angle(time=t, shell=s)
        kper_kpar = self.parameters['kper / kpar']
        kxx_kpar = numpy.cos(theta)**2 + kper_kpar*numpy.sin(theta)**2
        kxx = mfp * kxx_kpar * v / 3.0
        ux = self.umag(time=t, shell=s)
        return kxx / ux

    def average_energy(
        self,
        time=None,
        shell=None,
        species=None,
        energy=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the average energy.

        This function first approximates the integral of v²f(v) with respect to
        v, where v is the particle velocity and f(v) is the isotropic
        distribution, then computes the energy density, divides that result by
        the intermediate result computed here, and finally divides by 2 to
        represent the kinetic energy. In MKS, for example, the particle
        velocities have the unit m/s and the isotropic distribution has the unit
        s³/m⁶, giving the intermediate value a unit of 1/m³. Dividing the energy
        density by this intermediate quantity yields a quantity with the unit J.
        Similar dimensional analysis yields a result of erg in cgs.
        """
        v = self.dataset.v
        dv = numpy.gradient(v, axis=-1)
        # NOTE: This method does not attempt to subscript the isotropic
        # distribution by time or energy in case downstream code needs to
        # interpolate over them. It subscripts by shell because it assumes that
        # upstream code has reset `shell` as necessary for radial interpolation,
        # and it subscripts by species because interpolation over species is
        # meaningless.
        fiso = self._fiso(shell=shell, species=species)
        normalizer = numpy.sum(v**2 * fiso * dv, axis=-1)
        epsilon = self.energy_density(shell=shell, species=species)
        array = 0.5 * epsilon / normalizer
        unit = self.system['energy'].unit
        result = self._observe(array, assumptions, time=time, energy=energy)
        return result.withunit(unit)

    def energy_density(
        self,
        time=None,
        shell=None,
        species=None,
        energy=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the energy density.

        This function first approximates the integral of v⁴f(v) with respect to
        v, where v is the particle velocity and f(v) is the isotropic
        distribution, then multiplies that result by the species mass. In MKS,
        for example, the particle velocities have units of m/s and the isotropic
        distribution has units of s³/m⁶, giving the intermediate value units of
        (m/s)⁵s³/m⁶, or 1/m/s². Multiplying by the mass yields an expression
        with units of kg/m/s², or J/m³. Similar dimensional analysis yields a
        result of erg/cm³ in cgs.
        """
        v = self.dataset.v
        dv = numpy.gradient(v, axis=-1)
        # NOTE: This method does not attempt to subscript the isotropic
        # distribution by time or energy in case downstream code needs to
        # interpolate over them. It subscripts by shell because it assumes that
        # upstream code has reset `shell` as necessary for radial interpolation,
        # and it subscripts by species because interpolation over species is
        # meaningless.
        fiso = self._fiso(shell=shell, species=species)
        array = self.mass(species) * numpy.sum(v**4 * fiso * dv, axis=-1)
        unit = self.system['energy / volume'].unit
        result = self._observe(array, assumptions, time=time, energy=energy)
        return result.withunit(unit)

    def flux(
        self,
        time=None,
        shell=None,
        species=None,
        energy=None,
        mu=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the flux."""
        if self.dataset.hasflux:
            return self._observe(
                self.dataset.flux,
                assumptions,
                time=time,
                shell=shell,
                species=species,
                energy=energy,
            )
        # NOTE: This method does not attempt to subscript the isotropic
        # distribution by time or energy in case downstream code needs to
        # interpolate over them. It subscripts by shell because it assumes that
        # upstream code has reset `shell` as necessary for radial interpolation,
        # and it subscripts by species because interpolation over species is
        # meaningless. XXX: Maybe this method should check the assumptions
        # before deciding whether or not to subscript time and energy.
        fiso = self._fiso(shell=shell, species=species, mu=mu)
        e = self.energy()
        m = self.mass(species=species)
        g = physical.scalar(1.0, 'sr')
        array = 2 * fiso * e / (g * m**2)
        unit = self.system['flux'].unit
        result = array.withunit(unit)
        return self._observe(result, assumptions, time=time, energy=energy)

    def _fiso(
        self,
        time=None,
        shell=None,
        species=None,
        energy=None,
        mu=None,
    ) -> physical.Array:
        """Compute the isotropic distribution."""
        # XXX: If a bug related to energy interpolation appears, if may be due
        # to directly calling `energy` rather than "observing" energy within the
        # context of user assumptions. See notes at `mfp` and `rigidity`.
        if self.dataset.hasflux:
            flux = self._observe(
                self.dataset.flux,
                time=time,
                shell=shell,
                species=species,
                energy=energy,
            )
            m = self.mass(species=species)
            e = self.energy(energy=energy)
            # HACK: Explicitly construct an array in order to force the unit to
            # be the same. Otherwise, it will contain a solid-angle term.
            data = numpy.array((0.5 * flux * m**2) / e)
            unit = self.system['particle distribution'].unit
            return physical.array(data, unit=unit, axes=flux.axes)
        # NOTE: This method needs the full pitch-angle axis in order to
        # correctly compute the mean.
        f = self.f(
            time=time,
            shell=shell,
            species=species,
            energy=energy,
            mu=mu,
        )
        return numpy.mean(f, axis=-1)

    def fluence(
        self,
        time=None,
        shell=None,
        species=None,
        energy=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute the fluence."""
        # NOTE:
        # - This method does not pass time to `flux` so that it can compute the
        #   sum over the full time axis after observing `flux * dt`. However, it
        #   allows `flux` to subscript and interpolate over energy because
        #   energy is unaffected by the sum.
        # - This method observes the partial result, `flux * dt`, before
        #   indexing and summing over the time axis in order to allow the
        #   observing logic to interpolate to a radius if necessary.
        flux = self.flux(shell=shell, species=species, energy=energy)
        t = self.axes['time']
        it = t.indices
        tlo = self.time(time=it)
        thi = self.time(time=it.shift(-1, floor=0))
        dt = physical.array(
            numpy.array(tlo) - numpy.array(thi),
            unit=tlo.unit,
            axes={'time': t},
        )
        df = flux * dt
        observed = self._observe(df, assumptions, time=time)
        index = _create_slice_from(time)
        return numpy.sum(observed[index], axis=0)

    def _slice_fluence_df(self, time, array):
        """Subscript `array` with an appropriate slice along the time axis."""
        if isinstance(time, slice):
            return array[time]
        if isinstance(time, typing.SupportsIndex):
            return array[slice(int(time))]
        return array[:]

    def intflux(
        self,
        time=None,
        shell=None,
        species=None,
        assumptions: typing.Optional[dict]=None,
    ) -> physical.Array:
        """Compute intergral flux."""
        # NOTE: This method does not attempt to subscript flux by time in case
        # downstream code needs to interpolate over it. It does not attempt to
        # subscript flux by energy because it needs to integrate over that axis.
        # It subscripts by shell because it assumes that upstream code has reset
        # `shell` as necessary for radial interpolation, and it subscripts by
        # species because interpolation over species is meaningless.
        flux = self.flux(shell=shell, species=species)
        e = self.energy()
        minimum_energy = assumptions.pop('minimum_energy') or slice(None)
        intflux = self._integrate_flux(flux, e, minimum_energy)
        return self._observe(intflux, assumptions, time=time)

    def _integrate_flux(
        self,
        flux: physical.Array,
        energy: physical.Array,
        minimum_energy: parameter.ArgumentType | slice,
    ) -> physical.Array:
        """Helper for `intflux`."""
        if isinstance(minimum_energy, slice):
            return self._integrate_flux_slice(flux, minimum_energy)
        return self._integrate_flux_values(flux, energy, minimum_energy)

    def _integrate_flux_slice(
        self,
        flux: physical.Array,
        bins: slice,
    ) -> physical.Array:
        energy = self.axes['energy']
        ie = energy.indices[bins]
        elo = self.energy(energy=ie)
        ehi = self.energy(energy=ie.shift(-1, floor=0))
        de = physical.array(
            numpy.array(elo) - numpy.array(ehi),
            unit=elo.unit,
            axes={'energy': energy},
        )
        df = flux * de
        data = numpy.cumsum(df.data.array, axis=-1)
        extra = physical.axes({'minimum energy': elo})
        axes = flux.axes.without('energy', strict=True) | extra
        return physical.array(data, unit=df.unit, axes=axes)

    def _integrate_flux_values(
        self,
        flux: physical.Array,
        energy: physical.Array,
        minimum_energy: parameter.ArgumentType | typing.Iterable,
    ) -> physical.Array:
        bins = container.wrap(minimum_energy)
        data = numpy.zeros((*flux.shape[:-1], len(bins)))
        for i, bound in enumerate(bins):
            m = max(float(bound.withunit(energy.unit)), sys.float_info.min)
            y, x = self._check_interp(flux, energy, m)
            data[..., i] = integrate.simpson(y, x=x)
        unit = flux.unit * energy.unit
        extra = physical.axes({'minimum energy': minimum_energy})
        axes = flux.axes.without('energy', strict=True) | extra
        return physical.array(data, unit=unit, axes=axes)

    def _check_interp(
        self,
        y: numpy.typing.ArrayLike,
        x: numpy.typing.ArrayLike,
        p: float,
    ) -> typing.Tuple[numpy.ndarray, numpy.ndarray]:
        """Interpolate `y` at `x = m` if necessary.

        This function uses an algorithm that interpolates `y` at `x = m` via a
        power law. If `m` is already in `x`, the interpolation leaves `y`
        unchanged. It was designed to support computations of integral flux and
        therefore currently only interpolates to a lower bound.

        This function converts both `x` and `y` to a ``numpy.ndarray`` before
        applying the algorithm.
        """
        xarr = numpy.array(x)
        yarr = numpy.array(y)
        if p < xarr.min():
            return yarr, xarr
        if xarr.ndim != 1:
            raise ValueError(
                "Expected a 1-D coordinate array"
            ) from None
        n = xarr.size
        m = yarr.shape[-1]
        if n != m:
            raise ValueError(
                f"The length of the final dimension of the interpolant ({m})"
                f" must equal the length of the coordinate array ({n})"
            ) from None
        i0, x0 = container.nearest(xarr, p, bound='upper')
        try:
            lnxc = numpy.log(xarr[i0+1] / x0)
        except IndexError as err:
            raise ValueError(
                f"The point {p} is outside the coordinate array {xarr}"
            ) from err
        lnfc = numpy.log(yarr[..., i0+1] / yarr[..., i0])
        beta = lnfc / lnxc
        base = numpy.full_like(beta, p / x0)
        xarr[i0] = p
        yarr[..., i0] *= numpy.power(base, beta)
        return yarr[..., i0:], xarr[i0:]

    def _observe(
        self,
        x: physical.Array,
        assumptions: typing.Optional[typing.Dict[str, typing.Any]]=None,
        **indices: physical.AxisType,
    ) -> physical.Array:
        # NOTE: The motivation for distinguishing `assumptions` from `indices`
        # (as opposed to passing both as, e.g., `**user`) is that the need to
        # interpolate over a canonical coordinate manifests as a key-value pair
        # for that coordinate in both the indices and assumptions.
        """Apply final data processing to an observable quantity."""
        if not isinstance(x, physical.Array):
            raise TypeError(
                f"Cannot observe non-array object of type {type(x)}"
            ) from None
        if not assumptions:
            return self._subscript(x, indices)
        return self._interpolate(x, indices, assumptions)

    def _subscript(
        self,
        x: physical.Array,
        indices: typing.Dict[str, physical.AxisType],
    ) -> physical.Array:
        """Subscript the array based on user indices."""
        # Each value represents the user-provided indices, if available, or a
        # full slice. The use of `user.indices.get(d) or slice(None)` instead of
        # `user.indices.get(d, slice(None))` forces a slice even if
        # `user.indices[d] == None`.
        axes = tuple(indices.get(d) or slice(None) for d in x.dimensions)
        return x[_normalize_indices(*axes)]

    def _interpolate(
        self,
        x: physical.Array,
        indices: typing.Dict[str, physical.AxisType],
        assumptions: typing.Dict[str, typing.Any]=None,
    ) -> physical.Array:
        """Interpolate the given array based on indices and assumptions."""
        interpolants = self._compute_interpolants(x, assumptions or {})
        y = _prepare_array(x, indices, assumptions)
        base = interpolate(y, interpolants)
        default = {
            # Each value represents the appropriate default index for the
            # corresponding dimension of `x`, based on the size of that
            # dimension.
            d: numeric.index.value(0) if len(a) == 1 else slice(None)
            for d, a in base.axes.items()
        }
        finished = [
            # Each entry represents an interpolated dimension of `x`.
            'shell'if d in reference.ARRAYS.aliases['radius'] else d
            for d in interpolants
        ]
        axes = tuple(
            # Each value represents the user-provided indices, if available, or
            # the default indices for the corresponding dimension of `x`. The
            # use of `user.indices.get(d) or i` instead of `user.indices.get(d,
            # i)` forces the default value even if `user.indices[d] == None`.
            i if d in finished else indices.get(d) or i
            for d, i in default.items()
        )
        return base[_normalize_indices(*axes)]

    def _compute_interpolants(
        self,
        x: physical.Array,
        assumptions: dict,
    ) -> dict:
        """Determine the coordinate axes over which to interpolate."""
        from_axis = self._get_axis_interpolant(x, assumptions)
        if 'shell' not in x.dimensions:
            # The rest of this method deals with radial interpolation, which
            # only applies when 'shell' is one of the target quantity's
            # dimensions.
            return from_axis
        from_grid = self._get_grid_interpolant(x, assumptions, 'radius')
        return {**from_axis, **from_grid}

    def _get_axis_interpolant(self, x: physical.Array, assumptions: dict):
        """Compute interpolation information from an axis object."""
        interpolants = {}
        dimensions = {d for d in x.dimensions if d in self.grid}
        for key in dimensions:
            axis = self.axes.get(key)
            idx = x.axes[key].indices
            value = assumptions.get(key)
            if value is not None and isinstance(axis, physical.Coordinates):
                m = quantity.measure(value).withunit(axis.unit)
                a = physical.vector(axis[idx])
                for v in m:
                    _check_interpolation_bounds(key, v, a)
                interpolants[key] = Interpolant(
                    targets=numpy.array(m),
                    basis=self.grid[key][idx],
                )
        return interpolants

    def _get_grid_interpolant(
        self,
        x: physical.Array,
        assumptions: dict,
        name: str,
    ) -> dict:
        """Compute interpolation information from a grid object."""
        interpolants = {}
        indices = []
        for dimension in ('time', 'shell'):
            try:
                idx = x.axes[dimension].indices
            except KeyError:
                idx = slice(None)
            indices.append(idx)
        c = self.grid[name][tuple(indices)]
        a = physical.tensor(c)
        try:
            key = next(
                alias for alias in reference.ARRAYS.aliases[name]
                if alias in assumptions
            )
        except StopIteration:
            return interpolants
        if value := assumptions.get(key):
            m = quantity.measure(value).withunit(str(self.system))
            for v in m:
                _check_interpolation_bounds(name, v, a)
            interpolants[key] = Interpolant(
                targets=numpy.array(m),
                basis=c,
            )
        return interpolants

    @property
    def dataset(self):
        """The available dataset arrays."""
        return self._dataset

    @property
    def constants(self):
        """The available physical constants."""
        return self._constants

    @property
    def parameters(self):
        """The available runtime parameter values."""
        return self._parameters

    @property
    def system(self):
        """The metric system."""
        return self._system

    @property
    def axes(self):
        """The available array-indexing objects."""
        return self._axes

    @property
    def grid(self):
        """The available coordinate-grid arrays."""
        return self._grid


def _prepare_array(
    x: physical.Array,
    indices: typing.Dict[str, physical.AxisType],
    assumptions: typing.Dict[str, typing.Any]=None,
) -> physical.Array:
    """Perform any necessary pre-interpolation tasks.
    
    This function serves as a catch-all for updating the data to be
    interpolated. It returns a copy of `x` regardless of whether it actually
    modifies the associated data.
    """
    if 'time' in assumptions and str(x.unit) in {'rad', 'deg'}:
        return _patch_periodic_time(x, indices['time'])
    return physical.array(x)


def _patch_periodic_time(x: physical.Array, indices: physical.AxisType):
    """Adjust values of a periodic quantity for interpolation over time.
    
    Notes
    -----
    This function was written to fix a bug when interpolating `phi` (the
    azimuthal coordinate) over time, specifically for nodes with `phi` values
    that crossed the periodic boundary from one time step to the next. It has
    not been testing for other scenarios.
    """
    subarray = numpy.array(x[indices, :])
    diff = subarray[1] - subarray[0]
    absdiff = numpy.abs(diff)
    threshold = numpy.pi if x.unit == 'rad' else 180.0
    if numpy.max(absdiff) < threshold:
        return x
    node = numpy.argmax(absdiff)
    array = numpy.array(x)
    index = int(indices[1])
    sign = -numpy.sign(diff[node])
    value = 2*numpy.pi if x.unit == 'rad' else 360.0
    array[index, node] += sign * value
    return physical.array(array, unit=x.unit, axes=x.axes)


def _normalize_indices(*args):
    """Normalize array-indexing objects."""
    if all(isinstance(v, numeric.index.Value) for v in args):
        return tuple(int(arg) for arg in args)
    return tuple(
        numeric.index.sequence([int(arg)])
        if isinstance(arg, numeric.index.Value) else arg
        for arg in args
    )


def _create_slice_from(arg):
    """Convert `arg` to an appropriate slice object."""
    if isinstance(arg, slice):
        return arg
    if isinstance(arg, typing.SupportsIndex):
        return slice(int(arg))
    return slice(None)


def _check_interpolation_bounds(
    dimension: str,
    value: measured.Object,
    reference: measured.Object,
) -> None:
    """Raise an exception if `value` is not bounded by `reference`."""
    a = numpy.min(reference)
    b = numpy.max(reference)
    if not a < value < b:
        sv = f"{float(value):6.4g} {str(value.unit)}"
        sa = f"{float(a):6.4g} {str(reference.unit)}"
        sb = f"{float(b):6.4g} {str(reference.unit)}"
        raise ValueError(
            f"Cannot interpolate {dimension!r} to {sv}"
            f". Value must be in [{sa}, {sb}]"
        ) from None


@etc.autostr
class Quantities(collections.abc.Mapping):
    """A collection of array-like observable quantities.

    Notes
    -----
    For each observable quantity, there is a public-facing property and an
    internal callback method, the latter of which becomes the `operator`
    property of the `~Context` instance from which this class will construct the
    corresponding `~Quantity` instance. The operator method calls a
    quantity-specific implementation that contains all the logic necessary for
    computing the appropriate array (possibly including calling other
    implementations), given user indices.
    """

    def __init__(
        self,
        source: pathlib.Path,
        config: pathlib.Path,
        system: metric.System,
    ) -> None:
        """Initialize this interface."""
        self._source = source
        self._config = config
        self._system = system
        self._dataset = None
        self._axes = None
        self._grid = None
        self._context = None
        self._parameters = None
        self._constants = None
        self._cache = {}
        self._coordinates = None
        self._defined = None
        self._functions = None

    def __len__(self):
        """Called for len(self)."""
        return len(self.defined)

    def __iter__(self):
        """Called for iter(self)."""
        return iter(self.defined)

    def __getitem__(self, __k: str) -> Quantity:
        """Retrieve an observable array by name."""
        if __k in self._cache:
            return self._cache[__k]
        built = self._build_observable(__k)
        self._cache[__k] = built
        return built

    def _build_observable(self, key: str, /):
        """Retrieve or create an observable quantity by name."""
        name = reference.OBSERVABLES.names.get(key, key)
        expression = symbolic.expression(name, raising='**')
        term = expression[0]
        result = self._get_observable(term.base)
        if len(expression) == 1:
            # We don't need to multiply or divide quantities.
            if term.exponent == 1:
                # We don't even need to raise this quantity to a power.
                return result
            return result ** term.exponent
        # We need to apply symbolic operations.
        this = result ** term.exponent
        if len(expression) > 1:
            for term in expression[1:]:
                result = self._get_observable(term.base)
                this *= result ** term.exponent
        return this

    def _get_observable(self, key: str):
        """Internal helper for building an observable quantity."""
        name = reference.OBSERVABLES.names.get(key, key)
        if name in self.defined:
            return getattr(self, name)
        raise KeyError(
            f"{key!r} is not an observable quantity"
        ) from None

    @property
    def time(self):
        """The output times."""
        return self._implement(
            self._time,
            self.dataset.time.unit,
            self.dataset.time.dimensions,
        )

    def _time(self, user: Arguments):
        """Callback method for time."""
        if 'time' in user:
            unit = self.dataset.time.unit
            data = numpy.array(user['time'].withunit(unit))
            return physical.array(
                data,
                unit=unit,
                axes=self.dataset.time.axes,
            )
        return self.functions.time(time=user.time)

    @property
    def energy(self):
        """The particle energies per nucleon."""
        return self._implement(
            self._energy,
            self.dataset.energy.unit,
            self.dataset.energy.dimensions,
        )

    def _energy(self, user: Arguments):
        """Callback method for particle energies per nucleon."""
        if 'energy' in user:
            unit = self.dataset.energy.unit
            data = numpy.array(user['energy'].withunit(unit))
            return physical.array(
                data,
                unit=unit,
                axes=self.dataset.energy.axes,
            )
        return self.functions.energy(energy=user.energy)

    @property
    def v(self):
        """The particle speed at each energy per nucleon."""
        return self._implement(
            self._v,
            self.dataset.v.unit,
            self.dataset.v.dimensions,
        )

    def _v(self, user: Arguments):
        """Callback method for equivalent particle speeds."""
        if 'energy' in user:
            energy = self._energy(user)
            mass = self.constants['mp'].asscalar
            return physical.array(numpy.sqrt(2 * energy / mass))
        return self.functions.v(energy=user.energy)

    @property
    def mu(self):
        """The particle pitch-angle cosines."""
        return self._implement(
            self._mu,
            self.dataset.mu.unit,
            self.dataset.mu.dimensions,
        )

    def _mu(self, user: Arguments):
        """Callback method for pitch-angle cosines."""
        if 'mu' in user:
            unit = self.dataset.mu.unit
            data = numpy.array(user['mu'].withunit(unit))
            return physical.array(
                data,
                unit=unit,
                axes=self.dataset.mu.axes,
            )
        return self.functions.mu(mu=user.mu)

    @property
    def mass(self):
        """The mass of each species."""
        return self._implement(
            self._mass,
            self.dataset.mass.unit,
            self.dataset.mass.dimensions,
        )

    def _mass(self, user: Arguments):
        """Callback method for species mass."""
        return self.functions.mass(species=user.species)

    @property
    def charge(self):
        """The charge of each species."""
        return self._implement(
            self._charge,
            self.dataset.charge.unit,
            self.dataset.charge.dimensions,
        )

    def _charge(self, user: Arguments):
        """Callback method for species charge."""
        return self.functions.charge(species=user.species)

    @property
    def r(self):
        """The radial observer coordinates."""
        return self._implement(
            self._r,
            self.dataset.r.unit,
            self.dataset.r.dimensions,
        )

    def _r(self, user: Arguments):
        """Callback method for radial coordinates."""
        return self.functions.r(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def theta(self):
        """The polar observer coordinates."""
        return self._implement(
            self._theta,
            self.dataset.theta.unit,
            self.dataset.theta.dimensions,
        )

    def _theta(self, user: Arguments):
        """Callback method for polar coordinates."""
        return self.functions.theta(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def phi(self):
        """The azimuthal observer coordinates."""
        return self._implement(
            self._phi,
            self.dataset.phi.unit,
            self.dataset.phi.dimensions,
        )

    def _phi(self, user: Arguments):
        """Callback method for azimuthal coordinates."""
        return self.functions.phi(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def br(self):
        """The magnetic-field radial component."""
        return self._implement(
            self._br,
            self.dataset.br.unit,
            self.dataset.br.dimensions,
        )

    def _br(self, user: Arguments):
        """Callback method for radial magnetic field."""
        return self.functions.br(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def btheta(self):
        """The magnetic-field polar component."""
        return self._implement(
            self._btheta,
            self.dataset.btheta.unit,
            self.dataset.btheta.dimensions,
        )

    def _btheta(self, user: Arguments):
        """Callback method for polar magnetic field."""
        return self.functions.btheta(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def bphi(self):
        """The magnetic-field azimuthal component."""
        return self._implement(
            self._bphi,
            self.dataset.bphi.unit,
            self.dataset.bphi.dimensions,
        )

    def _bphi(self, user: Arguments):
        """Callback method for azimuthal magnetic field."""
        return self.functions.bphi(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def ur(self):
        """The velocity-field radial component."""
        return self._implement(
            self._ur,
            self.dataset.ur.unit,
            self.dataset.ur.dimensions,
        )

    def _ur(self, user: Arguments):
        """Callback method for radial velocity."""
        return self.functions.ur(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def utheta(self):
        """The velocity-field polar component."""
        return self._implement(
            self._utheta,
            self.dataset.utheta.unit,
            self.dataset.utheta.dimensions,
        )

    def _utheta(self, user: Arguments):
        """Callback method for polar velocity."""
        return self.functions.utheta(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def uphi(self):
        """The velocity-field azimuthal component."""
        return self._implement(
            self._uphi,
            self.dataset.uphi.unit,
            self.dataset.uphi.dimensions,
        )

    def _uphi(self, user: Arguments):
        """Callback method for azimuthal velocity."""
        return self.functions.uphi(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def rho(self):
        """The plasma density."""
        return self._implement(
            self._rho,
            self.dataset.rho.unit,
            self.dataset.rho.dimensions,
        )

    def _rho(self, user: Arguments):
        """Callback method for plasma density."""
        return self.functions.rho(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def mfp(self):
        """The scattering mean free path.
        
        Notes
        -----
        EPREM datasets generated on or after 11 Mar 2025 may contain the
        particle mean free path. In those cases, this property will prefer the
        dataset array. Otherwise, it will call an appropriate function to
        compute the equivalent mean free path.
        """
        if self.dataset.hasmfp:
            return self._implement(
                self._mfp,
                self.dataset.mfp.unit,
                self.dataset.mfp.dimensions,
            )
        unit = self.system['length'].unit
        dimensions = self.r.dimensions | self.rigidity.dimensions
        return self._implement(self._mfp, unit, dimensions)

    def _mfp(self, user: Arguments):
        """Callback method for mean free path."""
        if self.dataset.hasmfp:
            return self.functions.mfp(
                time=user.time,
                shell=user.shell,
                species=user.species,
                energy=user.energy,
                assumptions=user.assumptions,
            )
        if self.parameters.get('mfpInverseB', 0) == 1:
            return self.functions.mfp_B(
                time=user.time,
                shell=user.shell,
                species=user.species,
                energy=user.energy,
                assumptions=user.assumptions,
            )
        return self.functions.mfp_r(
            time=user.time,
            shell=user.shell,
            species=user.species,
            energy=user.energy,
            assumptions=user.assumptions,
        )

    @property
    def f(self):
        """The full particle distribution.
        
        Notes
        -----
        EPREM datasets generated with `outputFlux=1` contain particle flux in
        place of this quantity. In those cases, the observable quantity will
        have the correct unit and dimensions, but calling its operator will
        unconditionally return `None`.
        """
        if self.dataset.hasdist:
            return self._implement(
                self._f,
                self.dataset.f.unit,
                self.dataset.f.dimensions,
            )
        unit = self.system['particle distribution'].unit
        dimensions = self.dataset.flux.dimensions | 'mu'
        return self._implement(self._f, unit, dimensions)

    def _f(self, user: Arguments):
        """Callback method for particle distribution."""
        if self.dataset.hasdist:
            return self.functions.f(
                time=user.time,
                shell=user.shell,
                species=user.species,
                energy=user.energy,
                mu=user.mu,
                assumptions=user.assumptions,
            )

    @property
    def x(self):
        """The x-axis observer coordinates given by x = r·sin(θ)cos(φ)."""
        return self._implement(
            self._x,
            self.r.unit,
            self.r.dimensions,
        )

    def _x(self, user: Arguments):
        """Callback method for x coordinates."""
        return self.functions.x(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def y(self):
        """The y-axis observer coordinates given by y = r·sin(θ)sin(φ)."""
        return self._implement(
            self._y,
            self.r.unit,
            self.r.dimensions,
        )

    def _y(self, user: Arguments):
        """Callback method for coordinates."""
        return self.functions.y(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def z(self):
        """The z-axis observer coordinates given by z = r·cos(θ)."""
        return self._implement(
            self._z,
            self.r.unit,
            self.r.dimensions,
        )

    def _z(self, user: Arguments):
        """Callback method for z coordinates."""
        return self.functions.z(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def bmag(self):
        """The magnetic-field magnitude, |B|."""
        return self._implement(
            self._bmag,
            self.br.unit,
            self.br.dimensions,
        )

    def _bmag(self, user: Arguments):
        """Callback method for magnetic-field magnitude."""
        return self.functions.bmag(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def umag(self):
        """The velocity-field magnitude, |U|."""
        return self._implement(
            self._umag,
            self.ur.unit,
            self.ur.dimensions,
        )

    def _umag(self, user: Arguments):
        """Callback method for velocity-field magnitude."""
        return self.functions.umag(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def angle(self):
        """The angle between the magnetic- and velocity-field vectors."""
        return self._implement(
            self._angle,
            self.system['plane angle'].unit,
            self.ur.dimensions,
        )

    def _angle(self, user: Arguments):
        """Callback method for flow angle."""
        return self.functions.angle(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def upara(self):
        """The velocity-field component parallel to the magnetic field."""
        return self._implement(
            self._upara,
            self.ur.unit,
            self.ur.dimensions,
        )

    def _upara(self, user: Arguments):
        """Callback method for parallel velocity."""
        return self.functions.upara(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def uperp(self):
        """The velocity-field magnitude perpendicular to the magnetic field."""
        return self._implement(
            self._uperp,
            self.ur.unit,
            self.ur.dimensions,
        )

    def _uperp(self, user: Arguments):
        """Callback method for perpendicular velocity."""
        return self.functions.uperp(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def divu(self):
        """The velocity-field divergence."""
        return self._implement(
            self._divu,
            1 / self.system['time'].unit,
            self.rho.dimensions,
        )

    def _divu(self, user: Arguments):
        """Callback method for flow divergence."""
        return self.functions.divu(
            time=user.time,
            shell=user.shell,
            assumptions=user.assumptions,
        )

    @property
    def rigidity(self):
        """The magnetic rigidity."""
        unit = self.system['momentum / charge'].unit
        dimensions = self.mass.dimensions | self.energy.dimensions
        return self._implement(
            self._rigidity,
            unit,
            dimensions,
        )

    def _rigidity(self, user: Arguments):
        """Callback method for rigidity."""
        return self.functions.rigidity(
            species=user.species,
            energy=user.energy,
            assumptions=user.assumptions,
        )

    @property
    def ar(self):
        """The theoretical acceleration rate."""
        unit = 1 / self.system['time'].unit
        return self._implement(
            self._ar,
            unit,
            self.mfp.dimensions,
        )

    def _ar(self, user: Arguments):
        """Callback method for acceleration rate."""
        return self.functions.ar(
            time=user.time,
            shell=user.shell,
            species=user.species,
            energy=user.energy,
            assumptions=user.assumptions,
        )

    @property
    def average_energy(self):
        """The average particle-distribution energy."""
        unit = self.energy.unit
        removed = self.energy.dimensions | self.mu.dimensions
        dimensions = self.f.dimensions - removed
        return self._implement(
            self._average_energy,
            unit,
            dimensions,
        )

    def _average_energy(self, user: Arguments):
        """Callback method for average energy."""
        return self.functions.average_energy(
            time=user.time,
            shell=user.shell,
            species=user.species,
            energy=user.energy,
            assumptions=user.assumptions,
        )

    @property
    def energy_density(self):
        """The particle-distribution energy density."""
        unit = self.system['energy density'].unit
        return self._implement(
            self._energy_density,
            unit,
            self.average_energy.dimensions,
        )

    def _energy_density(self, user: Arguments):
        """Callback method for energy density."""
        return self.functions.energy_density(
            time=user.time,
            shell=user.shell,
            species=user.species,
            energy=user.energy,
            assumptions=user.assumptions,
        )

    @property
    def flux(self):
        """The differential energy flux of a distribution."""
        unit = self.system['flux'].unit
        if self.dataset.hasdist:
            dimensions = self.f.dimensions
        else:
            dimensions = self.f.dimensions - self.mu.dimensions
        return self._implement(
            self._flux,
            unit,
            dimensions,
        )

    def _flux(self, user: Arguments):
        """Callback method for particle flux."""
        return self.functions.flux(
            time=user.time,
            shell=user.shell,
            species=user.species,
            energy=user.energy,
            mu=user.mu,
            assumptions=user.assumptions,
        )

    @property
    def fluence(self):
        """The particle-distribution fluence."""
        unit = self.system['fluence'].unit
        return self._implement(
            self._fluence,
            unit,
            self.flux.dimensions,
        )

    def _fluence(self, user: Arguments):
        """Callback method for particle fluence."""
        return self.functions.fluence(
            time=user.time,
            shell=user.shell,
            species=user.species,
            energy=user.energy,
            assumptions=user.assumptions,
        )

    @property
    def intflux(self):
        """The integral flux of a distribution above a given energy"""
        unit = self.system['integral flux'].unit
        dimensions = self.flux.dimensions.replace('energy', 'minimum energy')
        return self._implement(
            self._intflux,
            unit,
            dimensions,
        )

    def _intflux(self, user: Arguments):
        """Callback method for integral particle flux."""
        assumptions = dict(user.assumptions)
        if 'minimum energy' not in assumptions:
            assumptions['minimum energy'] == sys.float_info.min
        return self.functions.intflux(
            time=user.time,
            shell=user.shell,
            species=user.species,
            assumptions=assumptions,
        )

    def _implement(
        self,
        operator: typing.Callable[[Arguments], physical.Array],
        /,
        unit: metric.Unit,
        dimensions: numeric.Dimensions,
    ) -> Quantity:
        """Implement an operation."""
        context = Context(operator, self.axes, self.grid)
        return Quantity(context, unit, dimensions)

    @property
    def defined(self):
        """The names and aliases of all observing implementations."""
        if self._defined is None:
            self._defined = [
                alias
                for k in self.functions.definitions
                for alias in reference.OBSERVABLES.aliases.get(k, [k])
            ]
        return self._defined

    @property
    def functions(self):
        """Functions corresponding to observable quantities."""
        if self._functions is None:
            self._functions = Implementations(
                self.dataset,
                self.axes,
                self.coordinates,
                self.constants,
                self.parameters,
                self.system,
            )
        return self._functions

    @property
    def grid(self):
        """The dataset coordinate-like arrays."""
        if self._grid is None:
            self._grid = datafile.grid(self.source, self.system)
        return self._grid

    @property
    def axes(self):
        """The array-indexing objects."""
        if self._axes is None:
            self._axes = datafile.axes(self.source, system=self.system)
        return self._axes

    @property
    def constants(self):
        """Fundamental physical constants in the given metric system."""
        if self._constants is None:
            self._constants = universal.Constants(str(self.system))
        return self._constants

    @property
    def coordinates(self):
        """The physical axes and pseudo-axes."""
        if self._coordinates is None:
            coordinates = {
                'time': self.dataset.time,
                'energy': self.dataset.energy,
                'mu': self.dataset.mu,
                'radius': self.dataset.r,
            }
            self._coordinates = aliased.Mapping(
                coordinates,
                aliases=reference.ALIASES,
            )
        return self._coordinates

    @property
    def dataset(self):
        """A view of the underlying dataset."""
        if self._dataset is None:
            self._dataset = datafile.arrays(self.source, self.system)
        return self._dataset

    @property
    def system(self):
        """The metric system of these quantities."""
        return self._system

    @property
    def source(self):
        """The data source."""
        return self._source

    @property
    def parameters(self):
        """The available runtime quantities."""
        if self._parameters is None:
            self._parameters = parameter.interface(self.config)
        return self._parameters

    @property
    def config(self):
        """The runtime configuration file."""
        return self._config


class Interpolant(typing.NamedTuple):
    """Attributes necessary for interpolation."""
    basis: physical.Array
    targets: numpy.ndarray


def interpolate(
    array: physical.Array,
    interpolants: typing.Mapping[str, Interpolant],
) -> physical.Array:
    """Interpolate a physical array.

    Parameters
    ----------
    array : `~Array`
        The array object to interpolate.

    interpolants : mapping from string to `~Interpolant`
    """
    if not interpolants:
        # If there are no interpolants, return the array as-is. This is simply a
        # convenience so that calling code need not check whether `interpolants`
        # is empty before calling this function.
        return array
    workspace = None
    axes = array.axes.copy()
    for dimension, interpolant in interpolants.items():
        workspace = _update_interpolation(
            array,
            interpolant.basis,
            interpolant.targets,
            workspace=workspace,
        )
        s = measured.sequence(interpolant.targets, interpolant.basis.unit)
        if dimension in ARRAYS.aliases['radius']:
            axes = axes.replace('shell', radius=physical.Coordinates(s))
        else:
            axes = axes.replace(dimension, physical.Coordinates(s))
    return array.spawn(workspace, unit=array.unit, axes=axes)


def _update_interpolation(
    array: physical.Array,
    basis: numeric.Array,
    targets: numpy.ndarray,
    workspace: numpy.ndarray=None,
) -> numpy.ndarray:
    """Helper for `~interpolate`."""
    x = numpy.array(array) if workspace is None else workspace
    indices = [array.dimensions.index(d) for d in basis.dimensions]
    dst, src = zip(*enumerate(indices))
    reordered = numpy.moveaxis(x, src, dst)
    interpolated = _apply_interpolation(reordered, basis, targets)
    return numpy.moveaxis(interpolated, dst, src)


def _apply_interpolation(
    data: numpy.typing.ArrayLike,
    reference: numpy.typing.ArrayLike,
    targets: typing.Iterable[float],
) -> numpy.ndarray:
    """Interpolate `data` to target values with respect to `reference`.

    This function will always interpolate over the leading dimension.

    Parameters
    ----------
    data : array-like
        The array-like object to interpolate. This function will convert `data`
        to a `numpy.ndarray`.

    reference : array-like
        The values at which `data` is known. This function will convert
        `reference` to a `numpy.ndarray`.

    targets : iterable of float
        The value(s) to which to interpolate `data`.

    Returns
    -------
    `numpy.ndarray`
        An array with the same shape as `array`, except for the interpolated
        dimension, which will have the same length as the number of values
        `targets`.
    """
    darray = numpy.array(data)
    rarray = numpy.array(reference)
    interpolated = [
        _apply_interp1d(darray, rarray, target)
        for target in targets
    ]
    if rarray.ndim == 2:
        return numpy.swapaxes(interpolated, 0, 1)
    return numpy.array(interpolated)


def _apply_interp1d(
    array: numpy.ndarray,
    reference: numpy.ndarray,
    target: float,
) -> typing.List[float]:
    """Interpolate data to `target` along the leading axis."""
    if target in reference:
        idx = container.nearest(reference, target).index
        return array[idx]
    ndim = reference.ndim
    if ndim == 2:
        interps = [interp1d(x, y, axis=0) for x, y in zip(reference, array)]
        return numpy.array([interp(target) for interp in interps])
    if ndim != 1:
        raise physical._array.NDimError(
            f"The reference array may have 1 or 2 (not {ndim}) dimensions"
        ) from None
    if reference.size == 1:
        # This works because we are interpolating over the leading dimension by
        # definition. Squeezing does not necessarily work because there may be
        # other singular dimensions.
        return array[0]
    interp = interp1d(reference, array, axis=0)
    return interp(target)





import abc
import re
import typing


class Part(abc.ABC):
    """Base class for parts of a symbolic expression."""

    __slots__ = ()


_PartType = typing.TypeVar('_PartType', bound=Part)


class Match(typing.Generic[_PartType]):
    """An object that represents the result of a RE pattern match."""

    def __new__(cls, result, context):
        """Check argument types."""
        if not isinstance(result, Part):
            raise TypeError(
                f"Result must be a Part"
                f", not {type(result)}"
            ) from None
        if not isinstance(context, (re.Match, typing.Mapping)):
            raise TypeError(
                f"Context may be a Match object or a Mapping"
                f", not {type(context)}"
            )
        return super().__new__(cls)

    def __init__(
        self,
        result: _PartType,
        context: typing.Union[re.Match, typing.Mapping],
    ) -> None:
        self.result = result
        """The result of the match attempt."""
        self._context = self._set_context(context)

    def _set_context(self, user: typing.Any) -> dict:
        """Normalize the instance context from user input."""
        if isinstance(user, re.Match):
            return self._set_from_match(user)
        if isinstance(user, typing.Mapping):
            return self._set_from_mapping(user)

    def _set_from_match(self, match: re.Match):
        """Set the instance context from a match object."""
        return {
            'start': match.start(),
            'end': match.end(),
            'string': match.string,
        }

    def _set_from_mapping(self, mapping: typing.Mapping):
        """Set the instance context from a mapping."""
        attrs = (
            'start',
            'end',
            'string',
        )
        return {k: mapping.get(k) for k in attrs}

    @property
    def start(self) -> int:
        """The starting index in `string` of the match."""
        return self._context['start']

    @property
    def end(self) -> int:
        """The ending index in `string` of the match."""
        return self._context['end']

    @property
    def string(self) -> str:
        """The target string."""
        return self._context['string']

    @property
    def remainder(self) -> str:
        """The unparsed portion of `string` after `end`."""
        return self.string[self.end:]

    def __bool__(self) -> bool:
        """Always true, like `re.Match`."""
        return True

    def __str__(self) -> str:
        """A simplified representation of this object."""
        attrs = {
            'result': 'result',
            'context': '_context',
        }
        return ', '.join(f"{k}: {getattr(self, v)}" for k, v in attrs.items())


class Factory(typing.Protocol):
    """Abstract protocol class for symbolic factories."""

    @abc.abstractmethod
    def parse(self) -> Match: ...


@typing.runtime_checkable
class FactoryType(Factory, typing.Protocol):
    """Runtime protocol for symbolic factories."""





import typing

from .. import metric
from ._context import (
    Context,
    context_factory as context,
)
from ._object import (
    Object,
    ObjectT,
)
from ._types import (
    Interface,
    Type,
)
from . import _factories
from . import _units as units
from . import _convert as convert
from ._sequence import Sequence
from ._value import Value


__all__ = [
    'Context',
    'Interface',
    'Object',
    'Object',
    'ObjectT',
    'Sequence',
    'Type',
    'Value',
    'units',
    'context',
    'convert',
]


T = typing.TypeVar('T')


@typing.overload
def value(
    x: typing.Any,
    /,
    unit: typing.Optional[metric.UnitLike],
) -> Value:
    """Create a measured value from input, if possible."""

@typing.overload
def value(x: T, /, unit: metric.UnitLike=typing.Literal['1']) -> Value[T]:
    """Create a measured value with the given unit."""

@typing.overload
def value(x: Value[T], /) -> Value[T]:
    """Create a measured value from an existing measured value."""

def value(x, /, unit=None) -> Value:
    return _factories.value_factory(x, unit=unit)


@typing.overload
def sequence(
    x: typing.Any,
    /,
    unit: typing.Optional[metric.UnitLike],
) -> Sequence:
    """Create a measured sequence from input, if possible."""

@typing.overload
def sequence(x: T, /, unit: metric.UnitLike=typing.Literal['1']) -> Sequence[T]:
    """Create a measured sequence with the given unit."""

@typing.overload
def sequence(x: Sequence[T], /) -> Sequence[T]:
    """Create a measured sequence from an existing measured sequence."""

def sequence(x, /, unit=None):
    return _factories.sequence_factory(x, unit=unit)





"""
Internal support for unit conversions.
"""

import numbers
import typing

from .. import etc
from .. import symbolic
from . import _defined
from . import _exceptions
from . import _reference


class Conversion:
    """The result of a unit conversion."""

    def __init__(self, u0: str, u1: str, factor: float=1.0) -> None:
        self._u0 = u0
        self._u1 = u1
        self._factor = factor

    def __rtruediv__(self, other):
        """Called to convert this conversion to its inverse."""
        if other != 1:
            clsname = self.__class__.__qualname__.lower()
            raise ValueError(
                f"Cannot compute {other} / <{clsname}>"
                f"; use {other} / float(<{clsname}>)"
            ) from None
        return Conversion(self.u1, self.u0, 1.0 / self._factor)

    def __float__(self) -> float:
        """Reduce this instance to its numerical factor via float(self)."""
        if bool(self):
            return self._factor
        raise TypeError("Conversion is undefined") from None

    def __bool__(self):
        """True if this conversion exists."""
        return self._factor is not None

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        attrstr = f"{self.u0!r}, {self.u1!r}, {self._factor}"
        return f"{self.__class__.__qualname__}({attrstr})"

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return f"({self.u0!r} -> {self.u1!r}) == {float(self)}"

    @property
    def u0(self):
        """The unit from which to convert."""
        return self._u0

    @property
    def u1(self):
        """The unit to which to convert."""
        return self._u1


_conversions = {}
"""Internal collection of computed unit conversions.

This module-level collection is a mapping from unit pairs to their corresponding
`~Conversion` instances. It exists to preclude the need to repeatedly compute
a given conversion.
"""

def conversion_factory(u0: str, u1: str) -> Conversion:
    """Compute the conversion from `u0` to `u1`.

    Parameters
    ----------
    u0 : string
        The unit from which to convert.

    u1 : string
        The unit to which to convert.

    Returns
    -------
    `~Conversion`

    Raises
    ------
    `~UnitConversionError`
        All attempts to convert `u0` to `u1` failed.

    Notes
    -----
    - This function will store both the successful conversion and its inverse in
      an internal module-level `dict` so that subsequent requests for the same
      conversion or its inverse do not trigger redundant computation.
    """
    if available := _conversions.get((u0, u1)):
        return available
    c = _convert(u0, u1)
    _conversions[(u0, u1)] = c
    _conversions[(u1, u0)] = 1 / c
    return c


def _convert(u0: str, u1: str):
    """Attempt to compute the conversion from `u0` to `u1`."""
    converters = (
        _convert_as_strings,
        _convert_as_expressions,
    )
    for converter in converters:
        if factor := converter(u0, u1):
            return Conversion(u0, u1, factor)
    raise _exceptions.UnitConversionError(
        f"Cannot convert {u0!r} to {u1!r}"
    ) from None


def _convert_as_strings(u0: str, u1: str) -> typing.Optional[float]:
    """Attempt to convert `u0` to `u1` as strings."""
    checked = []
    if factor := _recursive_conversion(checked, u0, u1):
        return factor


def _recursive_conversion(
    checked: typing.List[str],
    u0: str,
    u1: str,
    scale: float=1.0,
) -> typing.Optional[float]:
    """Attempt to convert `u0` to `u1` as strings."""
    if u0 not in checked:
        checked += [u0]
        if factor := _standard_conversion(u0, u1, scale=scale):
            return factor
        conversions = _reference.CONVERSIONS.get_adjacencies(u0).items()
        for unit, weight in conversions:
            if value := _recursive_conversion(checked, unit, u1, scale=scale):
                return weight * value


def _standard_conversion(
    u0: str,
    u1: str,
    scale: float=1.0,
) -> typing.Optional[float]:
    """Attempt to convert `u0` to `u1` non-recursively.

    This function will attempt to compute the appropriate numerical
    conversion factor via one of the following strategies

    - looking up a defined conversion
    - computing the ratio of metric scale factors
    - transitively scaling through an intermediate conversion
    """
    if factor := _simple_conversion(u0, u1, scale=scale):
        return factor
    if factor := _rescale_conversion(u0, u1, scale=scale):
        return factor


def _convert_as_expressions(u0: str, u1: str):
    """Convert complex unit expressions term-by-term."""
    e0, e1 = (symbolic.expression(unit) for unit in (u0, u1))
    if e0 == e1:
        return 1.0
    terms = [
        term for term in (e0 / e1).terms
        if term.base not in _reference.UNITY
    ]
    if factor := _resolve_terms(terms):
        return factor
    if factor := _convert_by_dimensions(terms):
        return factor


def _simple_conversion(u0: str, u1: str, scale: float=1.0):
    """Attempt to compute a simple conversion from `u0` to `u1`.
    
    This method will attempt the following conversions, in the order listed:
    
    * the identity conversion (i.e., `u0 == u1`);
    * a defined conversion from `u0` to `u1`;
    * the metric ratio of `u1` to `u0` (e.g., `'km'` to `'m'`);

    If it does not succeed, it will return `None` in order to allow other
    methods an opportunity to convert `u0` to `u1`.
    """
    if u0 == u1:
        return scale
    if found := _search(u0, u1):
        return scale * found
    try:
        ratio = _defined.NamedUnit(u0) >> _defined.NamedUnit(u1)
    except (ValueError, _exceptions.UnitParsingError):
        return
    return scale * ratio


def _rescale_conversion(u0: str, u1: str, scale: float=1.0):
    """Attempt to convert `u0` to `u1` via an intermediate unit.

    This method will attempt the following conversions, in the order listed:

    - a defined conversion to `u1` from a unit that has the same base unit as,
      but different metric scale than, `u0` (see `~Conversion._rescale`);
    - the numerical inverse of the above process applied to the conversion from
      `u1` to `u0`;

    If it does not succeed, it will return `None` in order to allow other
    methods an opportunity to convert `u0` to `u1`.
    
    Notes
    -----
    It is possible for `u0` or `u1` to individually be in `CONVERSIONS.nodes`
    even when `(u0, u1)` is not in `CONVERSIONS`. For example, there are nodes
    for both 'min' (minute) and 'd' (day), each with a conversion to 's'
    (second), but there is no direct conversion from 'min' to 'd'.
    """
    if u0 not in _reference.CONVERSIONS.nodes:
        if computed := _rescale(u0, u1):
            return scale * computed
    if u1 not in _reference.CONVERSIONS.nodes:
        if computed := _rescale(u1, u0):
            return scale / computed


def _rescale(u0: str, u1: str):
    """Compute a new conversion after rescaling `u0`.
    
    This method will look for a unit, `ux`, in `~metric.CONVERSIONS` that has
    the same base unit as `u0`. If it finds one, it will attempt to convert `ux`
    to `u1`, and finally multiply the result by the relative magnitude of `u0`
    to `ux`. In other words, it attempts to compute `(u0 << ux) * (ux -> u1)` in
    place of `(u0 -> u1)`.

    For example, if the direct conversion from megajoules to ergs is not
    available but the direct conversions from megajoules to joules and from
    joule to ergs are, this function will attempt to compute `('MJ' -> 'erg') =
    ('MJ' -> 'J') * ('J' -> 'erg')`.
    """
    if not u0 in _reference.NAMED_UNITS:
        return
    n0 = _defined.NamedUnit(u0)
    for ux in _reference.CONVERSIONS.nodes:
        if ux in _reference.NAMED_UNITS:
            nx = _defined.NamedUnit(ux)
            if nx.base == n0.base:
                if found := _search(ux, u1):
                    return (nx << n0) * found


def _search(u0: str, u1: str):
    """Search the defined conversions.
    
    This method will first search for a defined conversion from `u0` to `u1`. If
    it can't find one, it will search for a defined conversion from an alias of
    `u0` to an alias of `u1`. See `~_get_aliases_of` for more information.
    """
    if (u0, u1) in _reference.CONVERSIONS:
        return _reference.CONVERSIONS.get_weight(u0, u1)
    starts = _get_aliases_of(u0)
    ends = _get_aliases_of(u1)
    for ux in starts:
        for uy in ends:
            if (ux, uy) in _reference.CONVERSIONS:
                return _reference.CONVERSIONS.get_weight(ux, uy)


def _get_aliases_of(unit: str):
    """Build a list of possible variations of this unit string.
    
    The aliases of `unit` comprise the given string, as well as the canonical
    name and symbol of the corresponding named unit, if one exists.
    """
    built = [unit]
    if unit in _reference.NAMED_UNITS:
        known = _defined.NamedUnit(unit)
        built.extend([known.symbol, known.name])
    return built


def _convert_by_dimensions(terms: typing.List[symbolic.Term]):
    """Attempt to compute a conversion via unit dimensions."""
    factor = 1.0
    decomposed = []
    for term in terms:
        if decomposition := _decompose_unit(term):
            scale, units = decomposition
            decomposed.extend(units)
            factor *= scale
    if symbolic.expression(decomposed) == '1':
        return factor
    cancelled = _cancel_terms(decomposed)
    reduced = _reduce_terms(cancelled)
    if result := _resolve_terms(reduced):
        return factor * result


def _decompose_unit(term: symbolic.Term):
    """Convert a unit into fundamental units, if possible."""
    exponent = term.exponent
    if reduction := _reduce_unit(term.base, exponent):
        return 1.0, reduction
    for system in _reference.SYSTEMS:
        norm = _defined.NamedUnit(term.base).norm[system]
        if reduction := _reduce_unit(norm.symbol, exponent):
            if factor := _convert_as_strings(term.base, norm.symbol):
                return factor**exponent, reduction


def _reduce_unit(base: str, exponent: numbers.Real):
    """Compute the equivalent expression in fundamental units, if possible."""
    if reduction := _defined.NamedUnit(base).reduce():
        return [
            symbolic.term(
                coefficient=reduction.scale**exponent,
                base=this.base,
                exponent=exponent*this.exponent,
            )
            for this in reduction.units
        ]


def _cancel_terms(terms: typing.List[symbolic.Term]):
    """Cancel out terms with equal magnitude and opposite exponent."""
    # TODO: Reduce redundancy with `_resolve_terms`.
    matched = []
    unmatched = terms.copy()
    for target in terms:
        if target not in matched:
            inverse_powers = [
                term for term in terms
                if term != target and term.exponent == -target.exponent
            ]
            for term in inverse_powers:
                if term.base == target.base:
                    for this in (target, term):
                        matched.append(this)
                        unmatched.remove(this)
    return unmatched


def _resolve_terms(terms: typing.List[symbolic.Term]):
    """Compute ratios of terms with comparable exponents, if possible."""
    # TODO: Reduce redundancy with `_cancel_terms`.
    if len(terms) <= 1:
        # We require at least two terms for a ratio.
        return
    factor = 1.0
    matched = []
    unmatched = terms.copy()
    for target in terms:
        if target not in matched:
            if match := _match_exponents(target, unmatched):
                value, term = match
                if term != target:
                    for this in (target, term):
                        matched.append(this)
                        unmatched.remove(this)
                else:
                    matched.append(target)
                    unmatched.remove(target)
                factor *= value
    if not unmatched:
        return factor


def _reduce_terms(terms: typing.List[symbolic.Term]):
    """Partially or fully cancel terms with equal bases."""
    return symbolic.expression(terms).terms


def _match_exponents(
    target: symbolic.Term,
    terms: typing.Iterable[symbolic.Term],
) -> typing.Optional[typing.Tuple[float, symbolic.Term]]:
    """Attempt to convert `target` to a term in `terms` by exponent.

    This function first checks whether `target` is a dimensionless reference
    unit and, if so, returns a conversion factor of 1.0 along with `target`. If
    that check fails, this function compares `target` to each term in `terms`
    that has an exponent with the same magnitude and opposite sign. If it can
    convert the base unit of `target` to the base unit of a term in `terms`, it
    will return the corresponding conversion factor and the matching term.
    """
    u0 = target.base
    dimensions = _defined.NamedUnit(u0).dimensions.values()
    if u0 in _reference.NAMED_UNITS and all(d == '1' for d in dimensions):
        return 1.0, target
    exponent = target.exponent
    inverse_powers = [
        term for term in terms
        if term != target and term.exponent == -exponent
    ]
    for term in inverse_powers:
        u1 = term.base
        if conversion := _convert_as_strings(u0, u1):
            return conversion ** exponent, term


@etc.autostr
class Converter:
    """Unit-conversion handler for a known physical quantity."""

    def __init__(
        self,
        unit: str,
        quantity: str,
        substitutions: typing.Dict[str, str],
    ) -> None:
        self._unit = substitutions.get(unit) or unit
        self._quantity = quantity
        self._substitutions = substitutions

    def to(self, target: str):
        """Compute the conversion from the current unit to `target`.
        
        Parameters
        ----------
        target : string
            The unit or metric system to which to convert the current unit.

        Returns
        -------
        `~Conversion`

        Notes
        -----
        This method proceeds as follows:
        
        1. Substitute the appropriate unit if `target` is a metric system.
        1. Immediately return a trivial conversion object (i.e., with a `factor`
           of 1.0) if the target unit and the current unit are equal.
        1. Attempt a brute-force conversion via the `~factory` function and, if
           successful, return the resulting conversion object.
        1. Raise an exception to alert the caller that the conversion is
           undefined.
        """
        unit = self._substitutions.get(target) or target
        if self.unit == unit:
            return Conversion(self.unit, unit)
        if result := conversion_factory(self.unit, unit):
            return result
        raise ValueError(
            f"Unknown conversion from {self.unit!r} to {unit!r}."
        ) from None

    def __str__(self) -> str:
        return f"{self.unit!r} [{self.quantity!r}]"

    @property
    def unit(self):
        """The unit from which to convert."""
        return self._unit

    @property
    def quantity(self):
        """The physical quantity of `unit`."""
        return self._quantity


_converters = {}
"""Internal collection of unit-conversion handlers.

This module-level collection is a mapping from unit-quantity pairs to their
correpsonding `~Converter` instances. It exists to preclude the existence of
redundant unit-conversion handlers.
"""

def converter_factory(unit: str, quantity: str) -> Converter:
    """Create a new unit-conversion handler.

    Parameters
    ----------
    unit : string
        The unit to be converted.

    quantity : string
        The physical quantity of `unit`.

    Returns
    -------
    `~Converter`
        The unit-conversion handler corresponding to `unit` and `quantity`.

    Raises
    ------
    `ValueError`
        The given quantity is not a known physical quantity.
    """
    if available := _converters.get((unit, quantity)):
        return available
    substitutions = _get_substitutions(quantity)
    c = Converter(unit, quantity, substitutions)
    _converters[(unit, quantity)] = c
    return c


def _get_substitutions(quantity: str):
    """Get appropriate metric-system unit substitutions for `~converter`.
    
    This function exists to handle edge cases, such as treating atomic mass
    number quantities as mass quantities when converting to the equivalent unit
    in a given metric system.
    
    As a specific example, suppose you have a mass-like quantity with unit
    'nuc'. Since the metric quantity of 'nuc' is 'mass number', not 'mass', the
    default unit substitution would be '1' for both 'mks' and 'cgs'. However,
    intuition suggests that converting a mass-like quantity to 'mks' or 'cgs'
    should produce a quantity with unit 'kg' or 'g', respectively. This function
    therefore handles the special case of 'mass number' by replacing the default
    substitutions with the substitutions for 'mass'.
    """
    if quantity.lower() in {'mass number', 'mass_number'}:
        return _defined.UNITS['mass']
    try:
        substitutions = _defined.UNITS[quantity]
    except KeyError as err:
        raise ValueError(
            f"Unknown quantity {quantity!r}"
        ) from err
    return substitutions


def convert(source: str, target: str, quantity: str=None):
    """Convert `source` to `target`."""
    if not isinstance(source, str):
        raise TypeError(
            f"Argument to source must be a string, not {type(source)}"
        ) from None
    if not isinstance(target, str):
        raise TypeError(
            f"Argument to target must be a string, not {type(target)}"
        ) from None
    if target in _reference.SYSTEMS:
        if source == '1':
            return conversion_factory(source, '1')
        if quantity is None:
            if _is_ambiguous(source, target):
                raise _exceptions.UnitConversionError(
                    f"Conversion from {source!r} to {target!r}"
                    " is ambiguous without knowledge of physical quantity"
                ) from None
            q = _defined.get_quantity(symbolic.expression(source))
            return converter_factory(source, quantity=str(q)).to(target)
        return converter_factory(source, quantity=quantity).to(target)
    return conversion_factory(source, target)


def _is_ambiguous(unit: str, system: str):
    """True if the conversion from `unit` to `system` is ambiguous.

    Notes
    -----
    - This function is an internal helper for `~convert`.
    - This function will return `False` if either there is exactly one defined
      conversion from `unit` in `system`, or `unit` is not a canonical unit in
      `system`. The latter case does not guarantee a successful conversion; it
      simply allows this function to exit early if it has no chance of
      identifying an ambiguous conversion.
    """
    units = _defined.EQUIVALENT['units']
    others = {s for s in _reference.SYSTEMS if s != system}
    expr = symbolic.expression(unit)
    for s in others:
        targets = {v[system] for v in units.values() if v[s] == expr}
        if len(targets) > 1:
            return True
    return False




"""
Support for defining and working with physical array axes.
"""

import contextlib
import numbers
import typing

import numpy
import numpy.typing

from .. import container
from .. import etc
from .. import measured
from .. import metric
from .. import numeric
from .. import quantity
from ..typehelp import Self


T = typing.TypeVar('T')


class AxisValueError(ValueError):
    """Error in axis argument value."""


class AxisTypeError(TypeError):
    """Error in axis argument type."""


@etc.autostr
class Axis(numeric.Object[T]):
    """A generalized array axis."""

    def __init__(self, reference: T) -> None:
        super().__init__(reference)
        self._indices = None
        self._length = None

    def index(self, *targets: T):
        """Compute appropriate axis indices."""
        indices = self.normalize(*targets)
        if not all(isinstance(arg, numbers.Integral) for arg in indices):
            raise TypeError(
                f"Not all target indices have integral type: {targets}"
            ) from None
        if index := self._out_of_bounds(indices):
            raise ValueError(
                f"Index {index} is out of bounds for axis"
                f" with size {len(self)}"
            ) from None
        if len(indices) == 1:
            index = indices[0]
            with contextlib.suppress(TypeError):
                return numeric.index.value(int(index))
        return numeric.index.sequence(indices)

    def normalize(self, *targets: T):
        if not targets:
            return range(len(self))
        if len(targets) == 1:
            target = targets[0]
            if isinstance(target, slice):
                return container.slice2range(target, stop=len(self))
            if isinstance(target, (list, tuple, numpy.ndarray)):
                return target
            if container.isiterable(target):
                return list(target)
            # If `targets` is a tuple with a single member at this point,
            # execution will fall through and the function will return a list
            # containing only that element.
        return list(targets)

    def _out_of_bounds(self, indices: typing.Iterable[typing.SupportsIndex]):
        """Return the first index outside the bounds of this axis."""
        for index in indices:
            if index < -len(self) or index > len(self)-1:
                return index

    def __contains__(self, __x):
        """Called for x in self."""
        if isinstance(__x, Axis) and len(__x) == 1:
            with contextlib.suppress(Exception):
                return self.index(__x) in self.indices
        return __x in self.data

    def __getitem__(self, __i):
        """Called for index-based access.

        This method will generate a new instance of the appropriate (sub)class
        from its reference array. If `__i` is not `:` or `...`, this method will
        populate the internal `_indices` attribute with the indices
        corresponding to `__i`. Otherwise, it will immediately return the new
        axis object, which will automatically populate `_indices` as necessary.
        Both cases are designed to preserve index information with regard to the
        original reference array.
        """
        axis = self.spawn(self.data[__i])
        if __i in (slice(None), ...):
            return axis
        idx = self.normalize(container.unwrap(__i))
        axis._indices = numeric.index.sequence(idx)
        return axis

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self.data)

    def __iter__(self):
        """Called for iter(self)."""
        return iter(self.data)

    def __eq__(self, other):
        """Called for self == other."""
        if isinstance(other, Axis):
            return numpy.array_equal(self.data, other.data)
        return NotImplemented

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self._as_string()

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        return self._as_string(
            prefix=f'{self.__class__.__qualname__}(',
            suffix=')',
        )

    def _as_string(self, prefix: str='', suffix: str='') -> str:
        """Create a string representation of this object."""
        data = self._format_data(prefix=prefix, suffix=suffix)
        return f"{prefix}{data}{suffix}"

    def _format_data(self, **kwargs):
        """Create an appropriate string to represent the data."""
        array = numpy.array(self.data)
        try:
            signed = '+' if any(i < 0 for i in array.flat) else '-'
        except TypeError:
            # If we get here, it's probably because some values are strings
            signed = '-'
        return numpy.array2string(
            array,
            threshold=4,
            edgeitems=2,
            separator=', ',
            sign=signed,
            precision=3,
            floatmode='maxprec_equal',
            **kwargs
        )

    @property
    def indices(self) -> numeric.index.Sequence:
        """The corresponding index sequence."""
        if self._indices is None:
            self._indices = numeric.index.sequence(range(self.length))
        return self._indices

    @property
    def length(self) -> int:
        """The number of values in this axis."""
        if self._length is None:
            self._length = len(self.data)
        return self._length


AxisLike = typing.TypeVar('AxisLike', typing.Iterable[str], Axis)


class Points(Axis[numpy.typing.NDArray[numpy.uint]]):
    """The interface to an integral axis."""

    def __init__(self, data: numpy.typing.NDArray[numpy.uint]) -> None:
        super().__init__(data)

    def index(self, *targets: T):
        if not targets:
            return self.indices
        points = self.normalize(*targets)
        indices = []
        for point in points:
            if point not in self.data:
                raise AxisValueError(
                    f"This axis does not contain the point {point}"
                ) from None
            index = numpy.where(self.data == point)
            indices.append(index[0][0])
        if len(indices) == 1:
            return numeric.index.value(indices[0])
        return numeric.index.sequence(indices)

    @property
    def indices(self):
        if self._indices is None:
            self._indices = numeric.index.sequence(self.data)
        return self._indices


class Symbols(Axis[typing.List[str]]):
    """The interface to a symbolic axis."""

    def __init__(self, data: typing.List[str]) -> None:
        super().__init__(data)

    def __getitem__(self, index, /):
        if container.isiterable(index):
            return self.spawn([self.data[i] for i in index])
        return super().__getitem__(index)

    def index(self, *targets: T):
        if not targets:
            return self.indices
        if not all(isinstance(target, str) for target in targets):
            raise AxisTypeError(
                "All index targets must be strings"
            ) from None
        indices = [
            self.data.index(str(target))
            for target in targets
        ]
        if len(indices) == 1:
            return numeric.index.value(indices[0])
        return numeric.index.sequence(indices)


class Coordinates(Axis[measured.Sequence[numbers.Real]]):
    """The interface to a measured axis."""

    def __init__(self, data: measured.Sequence[numbers.Real]) -> None:
        super().__init__(data)

    def _as_string(self, prefix: str='', suffix: str='') -> str:
        """Create a string representation of this object."""
        data = self._format_data(prefix=prefix, suffix=suffix)
        return f"{prefix}{data}, {self.unit}{suffix}"

    def index(
        self,
        *targets: typing.Union[numbers.Real, str],
        closest: typing.Optional[str]=None,
    ) -> typing.Union[numeric.index.Value, numeric.index.Sequence]:
        if not targets:
            return self.indices
        this = self._measure(targets)
        if not this:
            raise AxisTypeError(f"Cannot measure {targets}") from None
        # At this point, we can assume that `this` is a measurement that we can
        # use to compute the indices of the target values. First, we need to
        # convert the measured values into the unit of this axis. That
        # conversion is straightfoward unless `targets` comprises only numbers,
        # in which case `this` will be unitless. In that case, we assume that
        # the caller meant to imply the use of the current unit (which may
        # actually be '1') and we therefore need to assign the current unit to
        # the measured values rather than convert them to the current unit.
        # - Cases:
        #   - targets are all numbers
        #     - the implicit unit is '1': We need to assign the current unit.
        #   - targets include a unit
        #     - the explicit unit is equivalent to the current unit: No change
        #       is necessary.
        #     - the explicit unit is not equivalent to the current unit: We need
        #       to convert them to the current unit.
        #     - the explicit unit is inconsistent with the current unit: This
        #       will raise an exception during conversion.
        #
        # Rather than separately handling each case, this method explicitly
        # assigns the measured unit to the measured value unless the measured
        # unit is '1', in which case it explicitly assigns the current unit. It
        # then explicitly converts the result to the current unit.
        # - Thus
        #   - if the target unit is implicitly '1' and the current unit is also
        #     '1', we will end up re-assigning the correct unit to the measured
        #     values at the cost of creating an intermediate sequence object,
        #     but the subsequent conversion to the current unit will be a no-op
        #   - if the target unit is implicitly '1' and the current unit is not
        #     '1', we will correctly assign the current unit to the measured
        #     values and the subsequence conversion to the current unit will be
        #     a no-op
        #   - if the targets have an explicit unit, we will end up re-assigning
        #     that unit to the measured values at the cost of creating an
        #     intermediate sequence object before converting the measured values
        #     to the current unit
        #
        # - Determine the most appropriate unit.
        unit = this.unit if this.unit != '1' else self.unit
        # - Create a temporary sequence with the intermediate unit. The
        #   following line casts `this` to a `numpy.ndarray` in order to force
        #   the unit update.
        tmp = measured.sequence(numpy.array(this), unit=unit)
        # - Convert the temporary array to this coordinate's unit and extract a
        #   numpy array. If the unit of `targets` is inconsistent with our unit,
        #   this is when we'll find out.
        array = numpy.array(tmp.withunit(self.unit))
        # - Convert target values to indices in the reference array. This will
        #   first make sure each target is close enough to a value in the
        #   reference array to be considered "in" the array. That check is
        #   important because `container.nearest` will naively find the
        #   numerically closest value, which may not be useful. It also lets
        #   calling code determine how to handle values not contained in the
        #   reference array (e.g., by interpolation or extrapolation).
        indices = [self._compute_index(target, closest) for target in array]
        if len(indices) == 1:
            return numeric.index.value(indices[0])
        return numeric.index.sequence(indices)

    def _compute_index(self, target, closest):
        """Helper for `index`."""
        if numeric.data.isclose(self.data, target):
            return container.nearest(self.data, target).index
        if closest is None:
            raise AxisValueError(
                f"{self.data} does not contain {target}"
            ) from None
        if closest == 'lower':
            return container.nearest(self.data, target, 'upper').index
        if closest == 'upper':
            return container.nearest(self.data, target, 'lower').index
        raise ValueError(f"Unknown constraint {closest!r}") from None

    def _measure(self, targets: typing.Sequence):
        """Create a measured sequence, if possible."""
        if len(targets) == 1:
            target = targets[0]
            if isinstance(target, quantity.Measurement):
                return target
            if isinstance(target, measured.Object):
                return quantity.measure(target)
            if quantity.ismeasurable(target):
                return quantity.measure(target)
        with contextlib.suppress(quantity.ParsingTypeError):
            return quantity.measure(targets)

    def __eq__(self, other):
        if isinstance(other, Coordinates):
            return super().__eq__(other) and self.unit == other.unit
        return super().__eq__(other)

    def __measure__(self) -> quantity.Measurement:
        """Called for `~quantity.measure(self)`."""
        return quantity.measurement(self.data)

    def withunit(self: Self, new: metric.UnitLike, /) -> Self:
        """Convert this cooridinate to a new unit."""
        s = self.data.withunit(new)
        return self.spawn(s)

    @property
    def unit(self):
        """The metric unit of reference values."""
        return self.data.unit





import collections
import collections.abc
import fractions
import functools
import numbers
from operator import attrgetter
import typing

from .. import etc
from . import _operand
from . import _parser


S = typing.TypeVar('S', str, numbers.Number)


class Expression(collections.abc.Sequence):
    """An object representing a symbolic expression."""

    def __init__(self, terms: typing.List[_operand.Term]) -> None:
        self._terms = terms

    def __bool__(self) -> bool:
        """True if this expression is not empty."""
        return not self.isempty

    @property
    def isempty(self):
        """True if this expression contains no terms."""
        return not self.terms

    def __iter__(self) -> typing.Iterator[_operand.Term]:
        return iter(self.terms)

    def __len__(self) -> int:
        return len(self.terms)

    @typing.overload
    def __getitem__(self, index: typing.SupportsIndex) -> _operand.Term: ...

    @typing.overload
    def __getitem__(self, index: slice) -> typing.Iterable[_operand.Term]: ...

    def __getitem__(self, index):
        """Access terms via standard indexing."""
        if isinstance(index, typing.SupportsIndex):
            idx = int(index)
            if idx > len(self):
                raise IndexError(index)
            if idx < 0:
                idx += len(self)
            return self.terms[idx]
        if isinstance(index, slice):
            return self.terms[index]
        raise IndexError(
            f"Invalid index for expression: {index}"
        ) from None

    def __str__(self) -> str:
        """A simplified representation of this instance."""
        return self.format()

    def __repr__(self):
        """An unambiguous representation of this object."""
        return f"{self.__class__.__qualname__}({self})"

    def format(self, style: str=None, *, separator: str=' '):
        """Join symbolic terms into a string.

        Parameters
        ----------
        style : string, optional
            The style in which to format each term. See `~Operand.format` for
            available styles.

        separator : string, default=' '
            The string to place between terms.
        """
        formatted = (term.format(style=style) for term in self)
        return separator.join(formatted)

    def difference(self, other, symmetric: bool=False, split: bool=False):
        """Compute the difference between two expressions.
        
        Parameters
        ----------
        other
            The object with respect to which to compute the difference. If
            `other` is not a `~symbolic.Expression`, this method will convert it
            to one before proceeding.

        symmetric : bool, default=False
            If true, compute the symmetric difference between this expression
            and `other`.

        split : bool, default=False
            If true, return the one-sided differences in a `list`. The first
            element contains the terms in this expression that are not in
            `other`, and the second element contains the terms in `other` that
            are not in this expression.

        Notes
        -----
        The `split` keyword argument takes precedence over the `symmetric`
        keyword argument because the result of the former contains more
        imformation than the result of the latter. See Examples for a suggestion
        on converting a split result into a symmetric result.

        Examples
        --------
        Consider the following two expressions:
        
        >>> e0 = symbolic.expression('a * b')
        >>> e1 = symbolic.expression('a * c')

        Their formal (one-sided) difference is

        >>> e0.difference(e1)
        {symbolic.term(b)}

        Their formal symmetric difference is

        >>> e0.difference(e1, symmetric=True)
        {symbolic.term(b), symbolic.term(c)}

        Passing `split=True` produces a `list` of `set`s

        >>> e0.difference(e1, split=True)
        [{symbolic.term(b)}, {symbolic.term(c)}]

        To convert a split result into a symmetric result, simply compute the
        union of the former:

        >>> symmetric = e0.difference(e1, symmetric=True)
        >>> split = e0.difference(e1, split=True)
        >>> set.union(*split) == symmetric
        True
        """
        if not isinstance(other, Expression):
            other = expression_factory(other)
        s0 = set(self.terms)
        s1 = set(other.terms)
        if split:
            return [s0 - s1, s1 - s0]
        if symmetric:
            return s0 ^ s1
        return s0 - s1

    def __hash__(self):
        """Compute hash(self)."""
        return hash(tuple(self.terms))

    @property
    def terms(self):
        """The symbolic terms in this expression."""
        return self._terms

    def __eq__(self, other) -> bool:
        """True if two expressions have the same symbolic terms.

        This method defines two expressions as equal if they have equivalent
        lists of symbolic terms (a.k.a simple parts), regardless of order, after
        parsing. Two expressions with different numbers of terms are always
        false. If the expressions have the same number of terms, this method
        will sort the triples (first by base, then by exponent, and finally by
        coefficient) and compare the sorted lists. Two expressions are equal if
        and only if their sorted lists of terms are equal.

        If `other` is not an instance of this class, this method will first
        attempt to convert it.
        """
        if not isinstance(other, Expression):
            other = expression_factory(other)
        if len(self) != len(other):
            return False
        key = attrgetter('base', 'exponent', 'coefficient')
        return sorted(self, key=key) == sorted(other, key=key)

    def __mul__(self, other):
        """Called for self * other.

        This method implements multiplication between two expressions by
        reducing the exponents of terms with a common base. If `other` is a
        string, it will first attempt to convert it to an `Expression`.
        """
        that = other if isinstance(other, Expression) else expression_factory(other)
        if not that:
            return NotImplemented
        reduced = reduce(self, that)
        return expression_factory(reduced)

    def __rmul__(self, other: typing.Any):
        """Called for other * self."""
        return expression_factory(other).__mul__(self)

    def __truediv__(self, other):
        """Called for self / other.

        This method implements division between two expressions by raising all
        terms in `other` to -1, then reducing the exponents of terms with a
        common base. If `other` is a string, it will first attempt to convert it
        to an `Expression`.
        """
        that = other if isinstance(other, Expression) else expression_factory(other)
        if not that:
            return NotImplemented
        return expression_factory(reduce(self, [term ** -1 for term in that]))

    def __rtruediv__(self, other: typing.Any):
        """Called for other / self."""
        return expression_factory(other).__truediv__(self)

    def __pow__(self, exp: numbers.Real):
        """Called for self ** exp.

        This method implements exponentiation of an expression by raising all
        terms to the given power, then reducing exponents of terms with a common
        base. It will first attempt to convert `exp` to a float.
        """
        exp = float(exp)
        if not exp:
            return NotImplemented
        terms = [pow(term, exp) for term in self]
        return expression_factory(reduce(terms))

    def apply(self, update: typing.Callable):
        """Create a new expression by applying the given callable object.
        
        Parameters
        ----------
        update : callable
            The callable object that this method should use to update the base
            of each term in this expression.

        Returns
        -------
        `~symbolic.Expression`
        """
        bases = [update(term.base) for term in self]
        exponents = [term.exponent for term in self]
        result = bases[0] ** exponents[0]
        for base, exponent in zip(bases[1:], exponents[1:]):
            result *= base ** exponent
        return result


@typing.overload
def expression_factory(*args, **kwargs) -> Expression:
    """Create an expression from the given arguments, if possible."""

@typing.overload
def expression_factory(expression: S, /, **kwargs) -> Expression:
    """Create an expression from a single value.

    Parameters
    ----------
    expression : string or number
        The value to convert into an expression.

    **kwargs
        Keywords to pass to `~symbolic.Parser`.

    Examples
    --------
    Create a symbolic expression from a string that represents the result of
    multiplying `a^3/2` by `b`, dividing by `c^1/2`, and squaring the ratio:

    >>> symbolic.expression('(a^3/2 * b / c^1/2)^2')
    symbolic.Expression(a^3 b^2 c^-1)
    """

@typing.overload
def expression_factory(
    expression: typing.Iterable[S],
    /,
    **kwargs,
) -> Expression:
    """Create an expression from an iterable of values.

    Parameters
    ----------
    expression : iterable of strings or numbers
        The iterable object with which to initialize the new instance. All
        members must support conversion to a string in the form of a
        `~symbolic.Term`.

    **kwargs
        Keywords to pass to `~symbolic.Parser`.

    Examples
    --------
    Create a symbolic expression from a list of the individual string terms in
    the result of multiplying `a^3/2` by `b`, dividing by `c^1/2`, and squaring
    the ratio:

    >>> symbolic.expression(['a^3', 'b', 'c^-1'])
    symbolic.Expression(a^3 b c^-1)
    """

@typing.overload
def expression_factory(expression: Expression, /) -> Expression:
    """Create an expression from an expression.

    This mode exists to support algorithms that don't know the type of argument
    until runtime. If the type is known, it is simpler to use the existing
    instance.

    Parameters
    ----------
    expression : `~symbolic.Expression`
        An existing instance of this class.

    Examples
    --------
    Create an instance from a string:

    >>> this = symbolic.expression('a * b / c')

    Pass the first instance to this class:

    >>> that = symbolic.expression(this)

    Both `this` and `that` represent the same expression...

    >>> this
    symbolic.expression(a b c^-1)
    >>> that
    symbolic.expression(a b c^-1)

    ...because they are the same object.

    >>> that is this
    True
    """

def expression_factory(this, **options):
    if isinstance(this, Expression):
        if not options:
            return this
        raise ValueError(
            f"Cannot update parsing options on an existing expression"
        ) from None
    string = standard(this, joiner='*')
    terms = _init_terms(string, _parser.Parser(**options))
    return Expression(terms)


def standard(
    this,
    missing: typing.Optional[str]=None,
    joiner: str='*',
) -> str:
    """Convert `this` to a standard format.
    
    Parameters
    ----------
    this : string or iterable
        The object to convert.

    missing : string, optional
        The value to return if `this` is null.

    joiner : string, default='*'
        The string token to use when joining parts of an iterable argument.

    See Also
    --------
    `~symbolic.Expression`: A class that represents one or more terms joined by
    symbolic operators and grouped by separator characters. Instances support
    multiplication and division with strings or other instances, and
    exponentiation by real numbers. Instantiation automatically calls this
    function.
    """
    if etc.isnull(this):
        return missing
    if isinstance(this, str):
        return this
    try:
        iter(this)
    except TypeError:
        return str(this)
    else:
        return joiner.join(f"({part})" for part in this)


def _init_terms(
    string: str,
    parser: _parser.Parser,
) -> typing.List[_operand.Term]:
    """Initialize terms for an expression by parsing `string`."""
    if string is None:
        return []
    terms = parser.parse(string)
    return reduce(terms)


def reduce(*groups: typing.Iterable[_operand.Term]):
    """Algebraically reduce terms with equal bases.

    Parameters
    ----------
    *groups : tuple of iterables
        One or more iterables of `~symbolic.Term` instances. If there are
        multiple groups, this method will combine all terms it finds in the full
        collection of groups.

    Notes
    -----
    This function will sort terms in order of ascending exponent, and
    alphabetically for equal exponents.
    """
    terms = [t for group in groups for t in group]
    reduced = {}
    for t in terms:
        if t.base in reduced:
            reduced[t.base]['coefficient'] *= t.coefficient
            reduced[t.base]['exponent'] += t.exponent
        else:
            attributes = {
                'coefficient': t.coefficient,
                'exponent': t.exponent,
            }
            reduced[t.base] = attributes
    fracs = [
        fractions.Fraction(v['coefficient'])
        for v in reduced.values()
    ]
    tmp = [
        _operand.term_factory(base=k, exponent=v['exponent'])
        for k, v in reduced.items()
        if k != '1' and v['exponent'] != 0
    ]
    # Sort: high to low in exponent, followed by alphabetic in base.
    variables = sorted(
        sorted(tmp, key=attrgetter('base')),
        key=attrgetter('exponent'),
        reverse=True,
    )
    c = functools.reduce(lambda x, y: x*y, fracs)
    constant = [_operand.term_factory(coefficient=c)]
    if not variables:
        return constant
    if c == 1:
        return variables
    return constant + variables





import contextlib
import numbers
import typing

import numpy

from .. import etc
from ..measured import Object as Measured
from .. import numeric
from .. import quantity
from .. import real
from ._scalar import Scalar
from ._vector import Vector
from ._tensor import Tensor
from ._array import Array
from ._object_factories import (
    scalar_factory,
    vector_factory,
    tensor_factory,
    array_factory,
)
from ._exceptions import PhysicalTypeError


T = typing.TypeVar('T')


def objects(operation):
    """Register `operation` as the implementation for all physical objects."""
    def wrapped(f):
        Scalar.implements(operation)(f)
        Vector.implements(operation)(f)
        Tensor.implements(operation)(f)
        Array.implements(operation)(f)
        return f
    return wrapped


@objects('eq')
def eq(a, b, /):
    """Compute a == b."""
    if isinstance(a, Measured) and isinstance(b, Measured):
        base = numeric.data.isequal(a.data, b.data) and a.unit == b.unit
        if isinstance(a, Array) and isinstance(b, Array):
            return base and a.axes == b.axes
        return base
    return False


@objects('ne')
def ne(a, b, /):
    """Compute a != b."""
    return not eq(a, b)


@objects('lt')
def lt(a, b, /):
    """Compute a < b."""
    return comparative(numeric.operators.lt, a, b)


@objects('le')
def le(a, b, /):
    """Compute a <= b."""
    return comparative(numeric.operators.le, a, b)


@objects('gt')
def gt(a, b, /):
    """Compute a > b."""
    return comparative(numeric.operators.gt, a, b)


@objects('ge')
def ge(a, b, /):
    """Compute a >= b."""
    return comparative(numeric.operators.ge, a, b)


@objects('abs')
def abs(a, /):
    """Compute abs(a)."""
    return unary(numeric.operators.abs, a)


@objects('pos')
def pos(a, /):
    """Compute +a."""
    return unary(numeric.operators.pos, a)


@objects('neg')
def neg(a, /):
    """Compute -a."""
    return unary(numeric.operators.neg, a)


@objects('add')
def add(a, b, /):
    """Compute a + b."""
    return additive(numeric.operators.add, a, b)


@objects('sub')
def sub(a, b, /):
    """Compute a - b."""
    return additive(numeric.operators.sub, a, b)


@objects('mul')
def mul(a, b, /):
    """Compute a * b."""
    return multiplicative(numeric.operators.mul, a, b)


@objects('truediv')
def truediv(a, b, /):
    """Compute a / b."""
    return multiplicative(numeric.operators.truediv, a, b)


@objects('floordiv')
def floordiv(a, b, /):
    """Compute a // b."""
    return multiplicative(numeric.operators.floordiv, a, b)


@objects('mod')
def mod(a, b, /):
    """Compute a % b."""
    return multiplicative(numeric.operators.mod, a, b)


@objects('pow')
def pow(a, b, mod: int=None):
    """Compute a ** b."""
    return exponential(numeric.operators.pow, a, b, mod=mod)


@Scalar.implements('round')
def round(a, /, ndigits: int=None):
    """Compute round(a[, ndigits])."""
    return unary(numeric.operators.round, a, ndigits=ndigits)


@Scalar.implements('floor')
def floor(a, /):
    """Compute floor(a)."""
    return unary(numeric.operators.floor, a)


@Scalar.implements('ceil')
def ceil(a, /):
    """Compute ceil(a)."""
    return unary(numeric.operators.ceil, a)


@Scalar.implements('trunc')
def trunc(a, /):
    """Compute trunc(a)."""
    return unary(numeric.operators.trunc, a)


@Scalar.implements('int')
def asint(a):
    """Compute int(self)."""
    return cast(int, a, typing.SupportsInt)


@Scalar.implements('float')
def asfloat(a):
    """Compute float(self)."""
    return cast(float, a, typing.SupportsFloat)


@Scalar.implements('complex')
def ascomplex(a):
    """Compute complex(self)."""
    return cast(complex, a, typing.SupportsComplex)


def cast(__type: T, a, protocol: typing.Optional[type]=None) -> T:
    """Convert `a` to `__type`, if possible."""
    if isinstance(a, Measured):
        if protocol and isinstance(a.data, protocol):
            return __type(a.data)
    raise PhysicalTypeError(
        f"Cannot convert {a} to {__type!r}"
    ) from None


def comparative(f: typing.Callable[..., T], a, b, /):
    """Compute a comparative binary operation."""
    if isinstance(a, Measured) and isinstance(b, Measured):
        if a.unit == b.unit:
            return f(a.data, b.data)
        raise ValueError(
            f"Cannot apply {f.__name__!r}"
            " to objects with different units."
        ) from None
    if isinstance(a, Measured) and isinstance(b, numbers.Real):
        return f(a.data, b)
    if isinstance(a, numbers.Real) and isinstance(b, Measured):
        return f(a, b.data)
    raise PhysicalTypeError(a)


def from_builtin(operation):
    """Process the result of a built-in operation."""
    def wrapped(f, *args, **kwargs):
        result = operation(f, *args, **kwargs)
        if isinstance(result, numeric.Result):
            return result.apply(_get_factory(*args))
        return result
    return wrapped


def from_numpy(operation):
    """Process the result of a numpy operation."""
    def wrapped(*args, **kwargs):
        result = operation(*args, **kwargs)
        factory = _get_factory(*args)
        if isinstance(result, numeric.Result):
            return result.apply(factory)
        if isinstance(result, (list, tuple)):
            converted = [
                x.apply(factory) if isinstance(x, numeric.Result) else x
                for x in result
            ]
            if isinstance(result, tuple):
                return tuple(converted)
            return converted
        return result
    return wrapped


def _get_factory(*operands):
    """Get the appropriate factory for creating a new object."""
    if any(isinstance(operand, Array) for operand in operands):
        return array_factory
    if any(isinstance(operand, Tensor) for operand in operands):
        return tensor_factory
    if any(isinstance(operand, Vector) for operand in operands):
        return vector_factory
    if any(isinstance(operand, Scalar) for operand in operands):
        return scalar_factory
    raise TypeError(
        f"Cannot determine appropriate factory from {operands}"
    ) from None


@from_builtin
def unary(f, a, /, **kwargs):
    """Compute a unary arithmetic operation."""
    if isinstance(a, Measured):
        try:
            unit = f(a.unit)
        except TypeError:
            unit = a.unit
        metadata = {'unit': unit}
        if isinstance(a, Array):
            metadata['axes'] = a.axes
        return numeric.Result(f(a.data, **kwargs), **metadata)
    raise PhysicalTypeError(a)


@from_builtin
def additive(f, a, b, /):
    """Compute an additive binary operation."""
    if isinstance(a, Measured):
        if a.unit != '1' and isinstance(b, (numbers.Number, numpy.number)):
            raise PhysicalTypeError(a, b)
    if isinstance(b, Measured):
        if b.unit != '1' and isinstance(a, (numbers.Number, numpy.number)):
            raise PhysicalTypeError(a, b)
    x = a if isinstance(a, Measured) else quantity.measure(a)
    y = b if isinstance(b, Measured) else quantity.measure(b)
    if not (isinstance(x, Measured) and isinstance(y, Measured)):
        raise PhysicalTypeError(x, y)
    opr = numeric.CALLABLES.find(f)
    msg = "Cannot {} objects with different {}."
    if isinstance(x, Array) and isinstance(y, Array):
        data = f(x.data, y.data)
        badattrs = []
        if x.unit == y.unit:
            if x.axes == y.axes:
                return array_factory(data, unit=x.unit, axes=x.axes)
            with contextlib.suppress(ValueError):
                axes = x.axes + y.axes
                return array_factory(data, unit=x.unit, axes=axes)
        if x.unit != y.unit:
            badattrs.append('units')
        if x.axes != y.axes:
            badattrs.append('bases')
        raise ValueError(msg.format(opr, etc.join(badattrs)))
    if isinstance(x, Array):
        if isinstance(y, Vector):
            raise TypeError(
                f"Cannot {opr} a vector to an array"
            ) from None
        data = f(x.data.array, y.data)
        if x.unit == y.unit:
            return array_factory(data, unit=x.unit, axes=x.axes)
        raise ValueError(msg.format(opr, 'units'))
    if isinstance(y, Array):
        if isinstance(x, Vector):
            raise TypeError(
                f"Cannot {opr} a vector to an array"
            ) from None
        data = f(x.data, y.data.array)
        if x.unit == y.unit:
            return array_factory(data, unit=x.unit, axes=y.axes)
        raise ValueError(msg.format(opr, 'units'))
    data = f(x.data, y.data)
    if x.unit == y.unit:
        return numeric.Result(data, unit=x.unit)
    raise ValueError(msg.format(opr, 'units'))


@from_builtin
def multiplicative(f, a, b):
    """Compute a multiplicative binary operation."""
    g = f if f == numeric.operators.mul else numeric.operators.truediv
    if isinstance(a, Measured) and isinstance(b, Measured):
        unit = g(a.unit, b.unit)
        if isinstance(a, Array) and isinstance(b, Array):
            data = f(a.data, b.data)
            axes = a.axes | b.axes
            return array_factory(data, unit=unit, axes=axes)
        if isinstance(a, Array):
            data = f(a.data.array, b.data)
            axes = a.axes
            return array_factory(data, unit=unit, axes=axes)
        if isinstance(b, Array):
            data = f(a.data, b.data.array)
            axes = b.axes
            return array_factory(data, unit=unit, axes=axes)
        return numeric.Result(f(a.data, b.data), unit=unit)
    if isinstance(a, Measured) and isinstance(b, numbers.Real):
        # If `b == 1`, both `a.unit * b` and `a.unit / b` will return `a.unit`.
        # If `b != 1`, neither `a.unit * b` nor `a.unit / b` is defined.
        # Therefore, we just set the new unit to `a.unit`.
        unit = a.unit
        if isinstance(a, Array):
            data = f(a.data.array, b)
            axes = a.axes
            return array_factory(data, unit=unit, axes=axes)
        return numeric.Result(f(a.data, b), unit=unit)
    if isinstance(a, numbers.Real) and isinstance(b, Measured):
        # We only want the unit or its inverse, depending on the operation,
        # regardless of the numerical value of `b`.
        unit = g(1, b.unit)
        if isinstance(b, Array):
            data = f(a, b.data.array)
            axes = b.axes
            return array_factory(data, unit=unit, axes=axes)
        return numeric.Result(f(a, b.data), unit=unit)
    x = a if isinstance(a, Measured) else quantity.measure(a)
    y = b if isinstance(b, Measured) else quantity.measure(b)
    if not (isinstance(x, Measured) and isinstance(y, Measured)):
        raise PhysicalTypeError(x, y)
    return multiplicative(f, x, y)


@from_builtin
def exponential(f, a, b, **kwargs):
    """Compute an exponential binary operation."""
    if isinstance(b, Measured) and not b.isunitless:
        raise ValueError(
            "Cannot compute a ** b unless b is unitless"
        ) from None
    if isinstance(a, Measured) and isinstance(b, Measured):
        if isinstance(b, Scalar):
            p = b.data
            unit = f(a.unit, p)
            if isinstance(a, Array):
                data = f(a.data.array, p, **kwargs)
                return array_factory(data, unit=unit, axes=a.axes)
            return numeric.Result(f(a.data, p, **kwargs), unit=unit)
        if not a.isunitless:
            raise ValueError(
                "Cannot compute a ** b when b is array-like"
                " unless both a and b are unitless"
            )
        # At this point, we know:
        # - both `a` and `b` are unitless
        # - `b` is array-like
        if isinstance(a, Array) and isinstance(b, Array):
            data = f(a.data.array, b.data.array, **kwargs)
            return array_factory(data, unit='1', axes=a.axes)
        if isinstance(a, Array):
            data = f(a.data.array, b.data, **kwargs)
            return array_factory(data, unit='1', axes=a.axes)
        if isinstance(b, Array):
            data = f(a.data, b.data.array, **kwargs)
            return array_factory(data, unit='1', axes=b.axes)
        return numeric.Result(f(a.data, b.data, **kwargs), unit='1')
    if isinstance(a, Measured) and isinstance(b, numbers.Real):
        unit = f(a.unit, b)
        if isinstance(a, Array):
            data = f(a.data.array, b, **kwargs)
            return array_factory(data, unit=unit, axes=a.axes)
        return numeric.Result(f(a.data, b, **kwargs), unit=unit)
    if isinstance(a, (numbers.Real, numpy.ndarray)) and isinstance(b, Measured):
        if isinstance(b, Array):
            return f(a, b.data.array, **kwargs)
        return f(a, b.data, **kwargs)
    raise PhysicalTypeError(a, b)


@from_numpy
@Tensor.implements(numpy.power)
@Vector.implements(numpy.power)
@Array.implements(numpy.power)
def power(a, b, **kwargs):
    """Called for numpy.power(a, b)."""
    if isinstance(b, (Vector, Tensor, Array)) and not b.isunitless:
        raise ValueError(
            "Cannot compute a ** b unless b is unitless"
        ) from None
    x = a.data if isinstance(a, Measured) else a
    y = b.data if isinstance(b, Measured) else b
    return numpy.power(x, y, **kwargs)


@objects(numpy.array_equal)
def array_equal(
    x: Measured[real.ValueType],
    y: Measured[real.ValueType],
) -> bool:
    return numpy.array_equal(numpy.array(x), numpy.array(y))


@objects(numpy.sin)
@from_numpy
def sin(x: Measured[real.ValueType]):
    """Compute the sin of `x`."""
    return _apply_trig(numpy.sin, x)


@objects(numpy.cos)
@from_numpy
def cos(x: Measured[real.ValueType]):
    """Compute the cosine of `x`."""
    return _apply_trig(numpy.cos, x)


@objects(numpy.tan)
@from_numpy
def tan(x: Measured[real.ValueType]):
    """Compute the tangent of `x`."""
    return _apply_trig(numpy.tan, x)


def _apply_trig(f, x: Measured[real.ValueType]):
    """Compute a trigonometric function of `x`."""
    if str(x.unit) in {'rad', 'deg'}:
        metadata = {'unit': '1'}
        if isinstance(x, Array):
            metadata['axes'] = x.axes
        return numeric.Result(f(x.data), **metadata)
    strf = f.__name__
    raise ValueError(
        f"Cannot compute {strf}(x)"
        f" when the unit of x is {str(x.unit)!r}"
    ) from None


@objects(numpy.sqrt)
@from_numpy
def sqrt(x: Measured[real.ValueType]):
    """Compute the square root of `x`."""
    data = numpy.sqrt(x.data)
    metadata = {'unit': x.unit ** 0.5}
    if isinstance(x, Array):
        metadata['axes'] = x.axes
    return numeric.Result(data, **metadata)


@objects(numpy.log)
@from_numpy
def log(x: Measured[real.ValueType]):
    """Compute the natural log of `x`."""
    return _apply_log(numpy.log, x)


@objects(numpy.log10)
@from_numpy
def log10(x: Measured[real.ValueType]):
    """Compute the base-10 log of `x`."""
    return _apply_log(numpy.log10, x)


@objects(numpy.log2)
@from_numpy
def log2(x: Measured[real.ValueType]):
    """Compute the base-2 log of `x`."""
    return _apply_log(numpy.log2, x)


@objects(numpy.log1p)
@from_numpy
def log1p(x: Measured[real.ValueType]):
    """Compute the natural log of `x + 1`."""
    return _apply_log(numpy.log1p, x)


def _apply_log(f, x: Measured[real.ValueType]):
    """Compute a logarithmic function of `x`."""
    if str(x.unit) == '1':
        metadata = {'unit': '1'}
        if isinstance(x, Array):
            metadata['axes'] = x.axes
        return numeric.Result(f(x.data), **metadata)
    strf = f.__name__
    raise ValueError(
        f"Cannot compute {strf}(x) when x is not unitless"
    ) from None


@objects(numpy.squeeze)
@from_numpy
def squeeze(x: Measured[real.ValueType], **kwargs):
    """Remove singular axes.

    Notes
    -----
    - Returning a scalar from a singular array is consistent with applying
      `numpy.squeeze` to a singular `numpy.ndarray`. For example,

      >>> numpy.squeeze(numpy.array([2])).ndim
      0
    """
    if isinstance(x, Scalar):
        return x
    data = numpy.squeeze(x.data, **kwargs)
    if isinstance(data, (numbers.Real)):
        return scalar_factory(data, unit=x.unit)
    if isinstance(data, numpy.ndarray) and data.ndim == 0:
        return scalar_factory(numpy.ravel(data)[0], unit=x.unit)
    metadata = {'unit': x.unit}
    if isinstance(x, Array):
        metadata['axes'] = {d: x.axes[d] for d in data.dimensions}
    return numeric.Result(data, **metadata)


@objects(numpy.min)
@from_numpy
def npmin(x: Measured[real.ValueType], **kwargs):
    """Compute the minimum value of `x`."""
    data = numpy.min(x.data, **kwargs)
    if isinstance(x, Scalar) or 'axis' not in kwargs:
        return scalar_factory(data, unit=x.unit)
    metadata = {'unit': x.unit}
    if isinstance(x, Array):
        metadata['axes'] = {d: x.axes[d] for d in data.dimensions}
    return numeric.Result(data, **metadata)


@objects(numpy.max)
@from_numpy
def npmax(x: Measured[real.ValueType], **kwargs):
    """Compute the maximum value of `x`."""
    data = numpy.max(x.data, **kwargs)
    if isinstance(x, Scalar) or 'axis' not in kwargs:
        return scalar_factory(data, unit=x.unit)
    metadata = {'unit': x.unit}
    if isinstance(x, Array):
        metadata['axes'] = {d: x.axes[d] for d in data.dimensions}
    return numeric.Result(data, **metadata)


@objects(numpy.mean)
@from_numpy
def mean(x: Measured[real.ValueType], **kwargs):
    """Compute the mean of `x`."""
    data = numpy.mean(x.data, **kwargs)
    if isinstance(x, Scalar) or 'axis' not in kwargs:
        return scalar_factory(data, unit=x.unit)
    metadata = {'unit': x.unit}
    if isinstance(x, Array):
        metadata['axes'] = {d: x.axes[d] for d in data.dimensions}
    return numeric.Result(data, **metadata)


@objects(numpy.sum)
@from_numpy
def sum(x: Measured[real.ValueType], **kwargs):
    """Compute the sum of `x`."""
    data = numpy.sum(x.data, **kwargs)
    if isinstance(x, Scalar) or 'axis' not in kwargs:
        return scalar_factory(data, unit=x.unit)
    metadata = {'unit': x.unit}
    if isinstance(x, Array):
        metadata['axes'] = {d: x.axes[d] for d in data.dimensions}
    return numeric.Result(data, **metadata)


@objects(numpy.cumsum)
@from_numpy
def cumsum(x: Measured[real.ValueType], **kwargs):
    """Compute the cumulative sum of `x`.

    Notes
    -----
    - Calling plain `numpy.cumsum` on a number, on a 1-D array, and without
      passing an argument to `axis` all produces a 1-D array. Calling the same
      function on an N-D array (N >= 1) and including an argument to `axis`
      produces an N-D array. That is why calling this implementation on a
      scalar, on a vector, and without passing an argument to `axis` produces a
      vector whereas calling it on a vector, tensor, or array and including an
      argument to `axis` produces an instance of the same object. Note that
      calling this implementation with `axis=0` or `axis=-1` on a scalar will
      produce a vector and doing so with any argument to `axis` will raise an
      exception. This behavior is also consistent with the plain `numpy.cumsum`
      implementation.
    """
    data = numpy.cumsum(x.data, **kwargs)
    if isinstance(x, (Scalar, Vector)) or 'axis' not in kwargs:
        return vector_factory(data, unit=x.unit)
    metadata = {'unit': x.unit}
    if isinstance(x, Array):
        metadata['axes'] = {d: x.axes[d] for d in data.dimensions}
    return numeric.Result(data, **metadata)


@objects(numpy.transpose)
@from_numpy
def transpose(x: Measured[real.ValueType], **kwargs):
    """Compute the transpose of `x`."""
    data = numpy.transpose(x.data, **kwargs)
    if isinstance(x, Scalar):
        return scalar_factory(numpy.ravel(data)[0], unit=x.unit)
    metadata = {'unit': x.unit}
    if isinstance(x, Array):
        metadata['axes'] = {d: x.axes[d] for d in data.dimensions}
    return numeric.Result(data, **metadata)


@objects(numpy.gradient)
@from_numpy
def gradient(x: Measured[real.ValueType], *args, **kwargs):
    """Compute the gradient of an array."""
    if isinstance(x, Scalar):
        return []
    if not args:
        return _apply_gradient(x.unit, x, **kwargs)
    diffs = []
    units = []
    for arg in args:
        if isinstance(arg, Scalar):
            diffs.append(float(arg))
            units.append(arg.unit)
        elif isinstance(arg, (Vector, Tensor, Array)):
            diffs.append(numpy.array(arg))
            units.append(arg.unit)
        elif isinstance(arg, numbers.Real):
            diffs.append(float(arg))
            units.append('1')
        else:
            diffs.append(numpy.array(arg))
            units.append('1')
    u0 = units[0]
    if all(unit == u0 for unit in units[1:]):
        return _apply_gradient(x.unit / u0, x, *diffs, **kwargs)
    raise TypeError("Inconsistent units in gradient coordinates") from None


def _apply_gradient(unit, x, *args, **kwargs):
    """Helper for custom implementation of `numpy.gradient`."""
    gradient = numpy.gradient(numpy.array(x), *args, **kwargs)
    metadata = {'unit': unit}
    if isinstance(x, Array):
        metadata['axes'] = x.axes
    if isinstance(gradient, (list, tuple)):
        return [numeric.Result(array, **metadata) for array in gradient]
    return numeric.Result(gradient, **metadata)


try:
    numpy.trapezoid
except AttributeError:
    numpy_trapezoid = numpy.trapz
else:
    numpy_trapezoid = numpy.trapezoid
@Array.implements(numpy_trapezoid)
@from_numpy
def trapezoid(x: Measured[real.ValueType], *args, **kwargs):
    """Integrate `x` via the composite trapezoidal rule."""
    if isinstance(x, Scalar):
        raise TypeError(
            "Cannot apply trapezoidal rule to a scalar object"
        ) from None
    data = numpy_trapezoid(x.data, *args, **kwargs)
    if isinstance(data, (numbers.Real, numpy.number)):
        return data
    metadata = {'unit': x.unit}
    if isinstance(x, Array):
        metadata['axes'] = {d: x.axes[d] for d in data.dimensions}
    return numeric.Result(data, **metadata)





"""
Factory functions for physical objects.

This module defines all factory functions because many (though not all) factory
functions require knowledge of types other than their target type.
"""

import numbers
import typing

import numpy

from .. import measured
from .. import quantity
from .. import real
from ._scalar import Scalar
from ._tensor import Tensor
from ._vector import Vector
from ._array import (
    Array,
    converter,
)
from ._axis import Coordinates
from ._axes import (
    Axes,
    axes_factory,
)


@Scalar.register.factory
def scalar_factory(x, /, unit=None) -> Scalar:
    """Internal factory function for `~Scalar` instances."""
    d, u = scalar_args(x, unit)
    ctx = measured.context(u, measured.convert.singular)
    return Scalar(d, ctx)


def scalar_args(x, unit, /):
    """Parse arguments to initialize `~Scalar`."""
    if isinstance(x, Scalar):
        return x.data, x.unit
    if isinstance(x, numbers.Real):
        return x, unit
    a = numpy.array(x)
    s = a.size
    if s != 1:
        raise TypeError(
            f"Cannot create a scalar from data with size {s}"
        ) from None
    data = a[0]
    if isinstance(x, Tensor):
        return data, x.unit
    return data, unit


@Vector.register.factory
def vector_factory(x, /, unit=None) -> Vector:
    """Internal factory for `~Vector` instances."""
    d, u = vector_args(x, unit)
    ctx = measured.context(u, measured.convert.arraylike)
    return Vector(d, ctx)


def vector_args(x, unit, /):
    """Parse arguments to initialize `~Vector`."""
    if isinstance(x, (Vector, Coordinates)):
        return x.data, x.unit
    if isinstance(x, (quantity.Measurement, Scalar)):
        y = [x.data] if isinstance(x, Scalar) else x
        return vector_args(numpy.asarray(y), x.unit)
    if isinstance(x, Tensor):
        return vector_args(numpy.array(x.data), x.unit)
    a = numpy.asarray(x)
    if a.ndim != 1:
        raise TypeError(
            "Vector data must be one-dimensional"
        ) from None
    return a, unit


@Tensor.register.factory
def tensor_factory(x, /, unit=None) -> Tensor:
    """Internal factory for `~Tensor` instances."""
    d, u = tensor_args(x, unit)
    ctx = measured.context(u, measured.convert.arraylike)
    return Tensor(d, ctx)


def tensor_args(x, unit, /):
    """Parse arguments to initialize `~Tensor`."""
    if isinstance(x, Array):
        return x.data.array, x.unit
    if isinstance(x, Tensor):
        return x.data, x.unit
    if isinstance(x, (quantity.Measurement, Scalar)):
        y = [x.data] if isinstance(x, Scalar) else x
        return tensor_args(numpy.asarray(y), x.unit)
    a = numpy.asarray(x)
    if a.ndim == 0:
        raise TypeError(
            "Tensor data must have at least one dimension"
        ) from None
    return a, unit


@Array.register.factory
def array_factory(x, /, unit=None, axes=None) -> Array:
    """Internal factory for `~Array` instances."""
    d, u, b = array_args(x, unit, axes)
    ctx = measured.context(u, converter)
    return Array(d, ctx, b)


def array_args(x, unit, axes, /):
    """Parse arguments to initialize `~Array`."""
    if isinstance(x, Array):
        return x.data, x.unit, x.axes
    dimensions = tuple(axes or ())
    data = real.array(array_data(x), *dimensions)
    axes = array_axes(data.shape, axes)
    if isinstance(x, measured.Object):
        return data, x.unit, axes
    return data, unit, axes


def array_shape(x) -> typing.Tuple[int]:
    """Compute the shape of `x`."""
    if isinstance(x, (Tensor, real.Array, numpy.ndarray)):
        return x.shape
    if isinstance(x, Vector):
        return (x.size,)
    if isinstance(x, Scalar):
        return (1,)
    if shape := getattr(x, 'shape', None):
        return shape
    raise TypeError(
        f"Cannot determine shape of {x}"
    ) from None


def array_data(x):
    """Extract data for a physical array.

    Notes
    -----
    The behavior of this function depends on the type of argument.

    - If the argument is an instance of `~real.Array`, this function will
      extract the internal `_object` attribute rather than the `array` property,
      in order to avoid prematurely loading a large dataset into memory.
    - If the argument is an instance of a subclass of `~measured.Object`, this
      function will extract the `data` property (converting a scalar-like object
      to 1-D array, if necessary).
    - If the argument is an instance of `numpy.ndarray`, this function will
      immediately return it.
    - If the argument is a logically N-D sequence, this function will convert to
      an N-D `numpy.ndarray`.
    - If the argument is a complex number, this function will convert it to a
      1-D real-valued `numpy.ndarray`.
    - This function will immediately return all other types of objects. Notably,
      it will NOT attempt to convert an arbitrary object to a `numpy.ndarray` in
      case doing so would prematurely load a large dataset into memory.
    """
    if isinstance(x, real.Array):
        return x._object
    if isinstance(x, (Tensor, Vector, measured.Sequence)):
        return x.data
    if isinstance(x, (Scalar, measured.Value)):
        return numpy.array([x.data])
    if isinstance(x, numpy.ndarray):
        return x
    if isinstance(x, numbers.Complex):
        return numpy.array([float(x)])
    if isinstance(x, typing.Sequence):
        return numpy.array(x)
    return x # <- NetCDF, HDF, etc.


def array_axes(shape, axes) -> Axes:
    """Create axes from arguments."""
    if isinstance(axes, typing.Mapping):
        return axes_factory(axes)
    try:
        axes = axes_factory(shape=shape, dimensions=axes)
    except (TypeError, ValueError) as err:
        raise TypeError(
            "Cannot create array axes"
            f" from shape={shape} and axes={axes}"
        ) from err
    return axes





"""
Support for working with observable EPREM quantities.
"""

import typing

from .. import metric
from .. import paths
from ._objects import (
    Array,
    Quantity,
)
from ._interfaces import Quantities

__all__ = [
    'Array',
    'Quantities',
    'Quantity',
    'quantities',
]


def quantities(
    source: paths.PathLike,
    config: paths.PathLike,
    system: typing.Optional[str]=None,
) -> Quantities:
    """Create a collection of observable arrays."""
    return Quantities(
        paths.fullpath(source, strict=True),
        paths.fullpath(config, strict=True),
        system=metric.system(system or 'mks')
    )





import typing

from . import _operand
from . import _operator


class ParsingError(Exception):
    """Base class for exceptions encountered during symbolic parsing."""

    def __init__(self, arg: typing.Any) -> None:
        self.arg = arg


class RatioError(ParsingError):
    """The string contains multiple '/' on a single level."""

    def __str__(self) -> str:
        return (
            f"The expression '{self.arg}' contains ambiguous '/'."
            f" Please refer to the NIST guidelines"
            f" (https://physics.nist.gov/cuu/Units/checklist.html)"
            f" for more information."
        )


class ProductError(ParsingError):
    """The string contains a '*' after a '/'."""

    def __str__(self) -> str:
        return (
            f"The expression '{self.arg}' contains an ambiguous '*'."
            f" Please group '*' in parentheses when following '/'."
        )


class ParsingValueError(ValueError):
    """Cannot create an expression from the given string."""
    pass


class Iteration:
    """An object that keeps track of parsing attributes."""

    __slots__ = ('string', 'operator', 'operand')

    def __init__(
        self,
        string: str,
        operator: _operator.Operator=None,
        operand: _operand.Operand=None,
    ) -> None:
        self.string = string
        self.operator = operator
        self.operand = operand

    @property
    def _attrs(self):
        """Internal mapping of current attribute values."""
        return {name: getattr(self, name) for name in self.__slots__}

    def copy(self):
        """Make a copy of this instance."""
        return type(self)(**self._attrs)

    def __str__(self):
        """A simplified representation of this object."""
        return ', '.join(f"{k}={v!r}" for k, v in self._attrs.items())

    def __repr__(self):
        """An unambiguous representation of this object."""
        return f"{self.__class__.__qualname__}({self})"


class Parser:
    """A tool for parsing symbolic expressions."""

    def __init__(
        self,
        multiply: str='*',
        divide: str='/',
        opening: str='(',
        closing: str=')',
        raising: str='^',
        operator_order: str='ignore',
    ) -> None:
        """
        Initialize a parser with fixed tokens.

        Parameters
        ----------
        multiply : string, default='*'
            The token that represents multiplication.

        divide : string, default='/'
            The token that represents division.

        opening : string, default='('
            The token that represents an opening separator.

        closing : string, default='('
            The token that represents a closing separator.

        raising : string, default='^'
            The token that represents raising to a power (exponentiation).

        operator_order : {'ignore', 'error'}
            Determines how the parser responds when operator order violates NIST
            guidelines. If set to `'ignore'` (default), it will treat operators
            independent of one another. If set to `'error'`, the parser will
            raise an exception based on the type of violation.
        """
        self.operands = _operand.OperandFactory(opening, closing, raising)
        self.operators = _operator.OperatorFactory(multiply, divide)
        self.parsers = (self.operands, self.operators)
        self.tokens = {
            'multiply': multiply,
            'divide': divide,
            'opening': opening,
            'closing': closing,
            'raising': raising,
        }
        self._operator_order = operator_order

    def parse(self, string: str):
        """Resolve the given string into individual terms."""
        operand = _operand.Operand(base=string)
        return self._resolve_operations(operand)

    def _resolve_operations(
        self,
        current: _operand.Operand,
    ) -> typing.List[_operand.Term]:
        """Separate a symbolic group into operators and operands."""
        operands = self._parse_operand(current)
        return [
            t for operand in operands
            for t in self._update_terms(operand)
        ] + [_operand.term_factory(coefficient=current.coefficient)]

    def _parse_operand(
        self,
        initial: _operand.Operand,
    ) -> typing.List[_operand.Operand]:
        """Resolve a general operand into simpler operands.

        This method parses known operators and operands from the initial operand
        while preserving nested groups in the latter. Calling code may then pass
        those nested groups back in for further parsing.
        """
        operands = []
        current = Iteration(initial.base)
        previous = current.copy()
        while current.string:
            current = self._get_operator(initial, current, previous)
            current = self._get_operand(initial, current)
            if new := self._compute_operand(current):
                operands.append(new)
            previous = current.copy()
            current = Iteration(previous.string)
        return operands

    def _get_operator(
        self,
        initial: _operand.Operand,
        current: Iteration,
        previous: Iteration,
    ) -> Iteration:
        """Attempt to parse an operator from the current string."""
        if parsed := self.operators.parse(current.string):
            current.operator = parsed.result
            if exception := self._operator_error(
                    current.operator,
                    previous.operator,
                ): raise exception(initial)
            current.string = parsed.remainder
        return current

    def _get_operand(
        self,
        initial: _operand.Operand,
        current: Iteration,
    ) -> Iteration:
        """Attempt to parse an operand from the current string."""
        if parsed := self.operands.parse(current.string):
            current.operand = parsed.result ** initial.exponent
            current.string = parsed.remainder
        return current

    def _compute_operand(self, current: Iteration):
        """Create a new operand from the current iteration."""
        if current.operand and current.operator:
            return self._evaluate(current.operator, current.operand)
        if current.operand:
            return current.operand
        if current.operator:
            raise ParsingValueError("Operator without operand")
        raise ParsingValueError("Failed to parse string")

    def _operator_error(
        self,
        current: _operator.Operator,
        previous: _operator.Operator,
    ) -> typing.Optional[typing.Type[ParsingError]]:
        """Check for known operator-related errors.
        
        This method checks for the following errors and returns the appropriate
        exception class if it finds one:

        - Multiple divisions on a single level (e.g., `'a / b / c'`), which
          results in a `RatioError`.

        - Multiplication after division on the same level (e.g., `'a / b * c'`),
          which results in a `ProductError`.

        Both of the examples shown above result in errors because they each
        introduce an ambiguous order of operations. Users can resolve the
        ambiguity by properly grouping terms in the expression. Continuing with
        the above examples, `'a / b / c'` should become `'(a / b) / c'` or `'a /
        (b / c)'`, and `'a / b * c'` should become `'(a / b) * c'` or `'a / (b *
        c)'`.
        """
        if self._operator_order == 'ignore':
            return
        if previous == 'divide':
            if current == 'divide':
                return RatioError
            if current == 'multiply':
                return ProductError

    def _evaluate(
        self,
        operator: _operator.Operator,
        operand: _operand.Operand,
    ) -> _operand.Operand:
        """Compute the effect of `operator` on `operand`."""
        if operator in {'multiply', 'identity'}:
            return operand
        if operator == 'divide':
            return operand ** -1
        if operator == 'sqrt':
            return operand ** 0.5
        raise ValueError(f"Unrecognized operator {operator!r}")

    def _update_terms(self, operand: _operand.Operand):
        """Store a new term or initiate further parsing."""
        # TODO: Consider extracting all coefficients, at least as separate
        # constant terms.
        if isinstance(operand, _operand.Term):
            return [operand]
        return self._resolve_operations(operand)


